class TriomaClass:

    def inspect(self, variable_names=None, tab: int = 0):
        """
        Prints specified variables of a class. If a variable is a class itself,
        calls this function recursively for the internal class. If variable_names is None,
        prints all variables.

        Args:
            instance: The class instance.
            variable_names (list of str, optional): Names of the variables to print. Prints all if None.
        """
        indent = "    " * tab
        for attr_name, attr_value in self.__dict__.items():
            if variable_names is None or attr_name.lower() == variable_names.lower():
                if type(attr_value) is TriomaClass:
                    tab += 1
                    print(
                        f"{indent}{attr_name} is a {type(attr_value)} class, printing its variables:"
                    )
                    TriomaClass.inspect(attr_value, variable_names, tab=tab)
                    tab -= 1
                else:
                    print(f"{indent}{attr_name}: {attr_value}")

    def update_attribute(self, attr_name: str, new_value: float):
        """
        Sets the specified attribute to a new value.

        Args:
            attr_name (str): The name of the attribute to set.
            new_value: The new value for the attribute.
        """
        if hasattr(self, attr_name):
            setattr(self, attr_name, new_value)
            if attr_name == "n_pipes":
                for attr, value in self.__dict__.items():
                    if isinstance(value, object) and hasattr(value, attr_name):

                        setattr(value, attr_name, new_value)
            return
        else:
            for attr, value in self.__dict__.items():
                if isinstance(value, object) and hasattr(value, attr_name):
                    setattr(value, attr_name, new_value)
                    return
        raise ValueError(
            f"'{attr_name}' is not an attribute of {self.__class__.__name__}"
        )



from TRIOMA.tools.Extractors.PAV import Component
from TRIOMA.tools.BreedingBlanket import BreedingBlanket
from TRIOMA.tools.Extractors.GasLiquidContactor import GLC
from TRIOMA.tools.TriomaClass import TriomaClass
import matplotlib.pyplot as plt
from typing import Union


class Circuit(TriomaClass):
    """
    Represent a circuit of components connected in series

    This class represents a circuit consisting of multiple components connected in series. It provides methods to update attributes, add components, calculate circuit efficiency, and plot the circuit.

    Attributes:
        components (list): A list of components in the circuit.
        closed (bool): A boolean indicating whether the circuit is a closed loop or not.

    Methods:
        update_attribute(attr_name, new_value):
            Updates the value of the specified attribute.

        add_component(component):
            Adds a component to the circuit.

        get_eff_circuit():
            Calculates the efficiency of the circuit.

        get_gains_and_losses():
            Calculates the gains and losses of the circuit.

        plot_circuit():
            Plots the circuit using matplotlib.

    Example usage:
        circuit = Circuit()
        circuit.add_component(component1)
        circuit.add_component(component2)
        circuit.get_eff_circuit()
        circuit.plot_circuit()
    """

    def __init__(
        self,
        components: list = None,
        closed: bool = False,
    ):
        vec_components = []
        if components is not None:
            for element in components:
                if isinstance(element, Union[Component, BreedingBlanket, GLC]):
                    vec_components.append(element)
                elif isinstance(element, Circuit):
                    for comp in element.components:
                        vec_components.append(comp)
                else:
                    raise ValueError("Invalid component type")
        self.components = vec_components
        self.closed = closed

    def add_component(
        self, component: Union["Component", "BreedingBlanket", "Circuit", "GLC"]
    ):
        """
        Adds a component to the circuit.

        Args:
            component (Component): The component to add.
        """
        if isinstance(component, Circuit):
            for comp in component.components:
                self.components.append(comp)
        else:
            self.components.append(component)

    def get_circuit_pumping_power(self):
        """
        Calculates the pumping power required for the circuit.

        Returns:
            float: The pumping power required for the circuit in W.
        """
        pumping_power = 0
        for component in self.components:
            if isinstance(component, Component):
                component.get_pressure_drop()
                component.get_pumping_power()
                pumping_power += component.pumping_power
        self.pumping_power = pumping_power
        return self.pumping_power

    def get_eff_circuit(self):
        """
        Calculates the efficiency of the circuit based on the components present.

        Returns:
            circtuit.eff (float): The efficiency of the circuit.

        Raises:
            None

        Example Usage:
            circuit.get_eff_circuit()

        """
        ind = None
        for i, component in enumerate(self.components):
            if isinstance(component, BreedingBlanket):
                ind = i
                break  # Assuming there's only one BreedingBlanket

        if ind is not None:
            # Move the element at index `ind` to the first position
            self.components = (
                [self.components[ind]]
                + self.components[:ind]
                + self.components[ind + 1 :]
            )

        for i, component in enumerate(self.components):
            if isinstance(component, GLC):
                component.get_c_out()
                if i != len(self.components) - 1:
                    component.connect_to_component(self.components[i + 1])
            if isinstance(component, Component):
                component.use_analytical_efficiency(p_out=component.p_out)
                component.outlet_c_comp()
            if i != len(self.components) - 1:
                component.connect_to_component(self.components[i + 1])
        eff_circuit = (
            self.components[1].c_in - self.components[-1].c_out
        ) / self.components[1].c_in
        self.eff = eff_circuit

    def get_gains_and_losses(self):
        """
        Calculates the gains and losses of the components in the circuit.

        Returns:
            circuit.extraction_perc (float): The extraction percentage of the circuit.
            circuit.loss_perc (float): The loss percentage of the circuit.

        """

        gains = 0
        losses = 0
        flag_bb = 0
        for i, component in enumerate(self.components):
            diff = component.c_in - component.c_out
            if isinstance(component, Component):
                if component.loss == False:
                    gains += diff
                else:
                    losses += diff
        for i, component in enumerate(self.components):

            if isinstance(component, BreedingBlanket):
                ind = i
                if flag_bb != 0:
                    print("There are more BB!")
                flag_bb = 1
        if ind != 0 and ind != len(self.components):
            eff_circuit = (
                self.components[ind + 1].c_in - self.components[ind - 1].c_out
            ) / self.components[ind + 1].c_in
            self.extraction_perc = gains / self.components[ind + 1].c_in / eff_circuit
            self.loss_perc = losses / self.components[ind + 1].c_in / eff_circuit
        elif ind == 0:
            eff_circuit = (
                self.components[ind + 1].c_in - self.components[-1].c_out
            ) / self.components[ind + 1].c_in
            self.extraction_perc = gains / self.components[ind + 1].c_in / eff_circuit
            self.loss_perc = losses / self.components[ind + 1].c_in / eff_circuit
        elif ind == len(self.components):
            eff_circuit = (
                self.components[0].c_in - self.components[ind - 1].c_out
            ) / self.components[0].c_in
            self.extraction_perc = gains / self.components[0].c_in / eff_circuit
            self.loss_perc = losses / self.components[0].c_in / eff_circuit
        self.eff = eff_circuit

    def plot_circuit(self):
        """
        Plot the circuit diagram for the components in the circuit.

        This function uses matplotlib to create a circuit diagram for the components in the circuit.
        Each component is represented by a rectangle with arrows indicating the flow direction.
        The color of the rectangle represents the position of the component in the circuit.

        Returns:
            None
        """
        from matplotlib.colors import LinearSegmentedColormap

        # Define the red-to-blue colormap
        red_to_blue = LinearSegmentedColormap.from_list("RedToBlue", ["red", "blue"])
        num_components = len(self.components)
        if num_components < 10:
            num_rows = 1
            num_columns = num_components
        else:
            num_rows = (num_components // 10) + (1 if num_components % 10 != 0 else 0)
            num_columns = 10

        fig, axs = plt.subplots(
            num_rows + 1, num_columns, figsize=(5 * num_columns, 4 * num_rows)
        )
        fig.subplots_adjust(hspace=0)
        # for component in self.components:
        #     component.plot_component()
        for i, component in enumerate(self.components):
            color = red_to_blue(i / num_components)
            if isinstance(component, Component):
                rectangle = plt.Rectangle(
                    (0.2, 0.3), 0.55, 0.4, edgecolor="black", facecolor=color, alpha=0.5
                )
                axs[i // num_columns, i % num_columns].add_patch(rectangle)
                # Arrow pointing to the left side of the rectangle
                axs[i // num_columns, i % num_columns].arrow(
                    0.0,
                    0.5,
                    0.1,
                    0,
                    head_width=0.05,
                    head_length=0.1,
                    fc="black",
                    ec="black",
                )
                # Arrow pointing out of the right side of the rectangle
                axs[i // num_columns, i % num_columns].arrow(
                    0.8,
                    0.5,
                    0.1,
                    0,
                    head_width=0.05,
                    head_length=0.1,
                    fc="black",
                    ec="black",
                )
                axs[i // num_columns, i % num_columns].set_aspect("equal")
                axs[i // num_columns, i % num_columns].set_xlim(0, 1)
                axs[i // num_columns, i % num_columns].set_ylim(0, 1)
                axs[i // num_columns, i % num_columns].text(
                    0.15,
                    0.3,
                    f"L={component.geometry.L:.3g} m",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.15,
                    0.4,
                    f"T={component.fluid.T:.6g}K",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.15,
                    0.6,
                    f"$c_{{in}}={component.c_in:.4g}  mol/m^3$",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.5,
                    0.7,
                    f"velocity={component.fluid.U0:.2g} m/s",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.5,
                    0.4,
                    f"eff={component.eff*100:.2g}%",
                    color="black",
                    ha="center",
                    va="center",
                )

                axs[i // num_columns, i % num_columns].text(
                    0.9,
                    0.3,
                    rf"$c_{{out}}={component.c_out:.4g} mol/m^3$",
                    color="black",
                    ha="center",
                    va="center",
                )
            elif isinstance(component, BreedingBlanket):

                rectangle = plt.Rectangle(
                    (0.2, 0.3),
                    0.55,
                    0.4,
                    edgecolor="black",
                    facecolor="green",
                    alpha=0.5,
                )
                axs[i // num_columns, i % num_columns].add_patch(rectangle)
                # Arrow pointing to the left side of the rectangle
                axs[i // num_columns, i % num_columns].arrow(
                    0.5,
                    0.7,
                    0,
                    0.1,
                    head_width=0.05,
                    head_length=0.1,
                    fc="black",
                    ec="black",
                )
                # Arrow pointing out of the right side of the rectangle
                axs[i // num_columns, i % num_columns].arrow(
                    0.5,
                    0.1,
                    0.0,
                    0.1,
                    head_width=0.05,
                    head_length=0.1,
                    fc="black",
                    ec="black",
                )
                axs[i // num_columns, i % num_columns].set_aspect("equal")
                axs[i // num_columns, i % num_columns].set_xlim(0, 1)
                axs[i // num_columns, i % num_columns].set_ylim(0, 1)
                if component.name is None:
                    axs[i // num_columns, i % num_columns].set_title("Component ")
                else:
                    axs[i // num_columns, i % num_columns].set_title(component.name)
                axs[i // num_columns, i % num_columns].text(
                    0.7,
                    0.8,
                    r"$T_o$=" + str(component.T_out) + " K",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.7,
                    0.2,
                    r"$T_i$=" + str(component.T_in) + " K",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.3,
                    0.2,
                    f"$c_i$={component.c_in:.4g} $mol/m^3$",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.5,
                    0.6,
                    f"Q={component.Q/1E6:.3g} MW",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.5,
                    0.4,
                    f"TBR={component.TBR:.3g}",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].text(
                    0.3,
                    0.8,
                    rf"$c_o$={component.c_out:.4g}$mol/m^3$",
                    color="black",
                    ha="center",
                    va="center",
                )
                axs[i // num_columns, i % num_columns].axis("off")
                # Display the plot
                fig.tight_layout()
            else:
                raise ValueError("Component not recognized")
        for row in axs:
            for ax in row:
                ax.axis("off")
                ax.set_ylim(0.2, 0.8)
        return fig

    def get_inventory(self, flag_an=True):
        """
        Calculates the inventory (in mol) of the circuit based on the components present.

        Returns:
            circuit.inv (float): The inventory of the circuit.

        Raises:
            None

        Example Usage:
            circuit.get_circuit_inventory()

        """
        inventory = 0
        for component in self.components:
            if isinstance(component, Component):
                component.get_inventory(flag_an=flag_an)
                inventory += component.inv
        self.inv = inventory

    def solve_circuit(self, tol=1e-6):
        """
        Solve the circuit by calculating the concentration of the components at the outlet.
        If the circuit is a closed loop, the concentration of the first component is set to the concentration of the last component until the stationary regime is reached.

        """
        err = 1
        flag = 0
        flag_bb = 0
        for i, component in enumerate(self.components):

            if isinstance(component, BreedingBlanket):
                ind = i
                if flag_bb != 0:
                    print("There are more BB!")
                flag_bb = 1
        while flag == 0:
            for i, component in enumerate(self.components):
                if isinstance(component, GLC):
                    component.get_c_out()
                    if i != len(self.components) - 1:
                        component.connect_to_component(self.components[i + 1])
                if isinstance(component, Component):
                    component.use_analytical_efficiency(p_out=component.p_out)
                    component.outlet_c_comp()
                    if i != len(self.components) - 1:
                        component.connect_to_component(self.components[i + 1])
                if isinstance(component, BreedingBlanket):
                    component.get_cout()
                    if i != len(self.components) - 1:
                        component.connect_to_component(self.components[i + 1])
            if self.closed == True:
                err = (
                    abs(self.components[0].c_in - self.components[-1].c_out)
                    / self.components[0].c_in
                )
                if err < tol:
                    flag = 1
            else:
                flag = 1
            self.components[-1].connect_to_component(self.components[0])

    def inspect_circuit(self, name=None):
        """
        Inspects the circuit components.

        Parameters:
            name (str, optional): The name of the component to inspect. If not provided, all components will be inspected.

        Returns:
            None

        """
        if name is None:
            for component in self.components:
                component.inspect()
        else:
            for component in self.components:
                if component.name == name:
                    component.inspect()

    def estimate_cost(self, metal_cost=0, fluid_cost=0):
        """
        Estimates the cost of the circuit.
        metal_cost: cost of the metal in $/m^3 in a vector ordered as the components
        fluid_cost: cost of the fluid in $/m^3 in a vector ordered as the components
        returns the cost of the component
        """
        cost_circuit = 0
        for i, component in enumerate(self.components):
            if isinstance(component, Component):
                component.estimate_cost(
                    metal_cost=metal_cost[i], fluid_cost=fluid_cost[i]
                )
                cost_circuit += component.cost
        self.cost = cost_circuit
        return self.cost



from TRIOMA.tools.TriomaClass import TriomaClass
from TRIOMA.tools.Extractors.PipeSubclasses import Fluid, Membrane
import TRIOMA.tools.correlations as corr
import TRIOMA.tools.Extractors.extractor as extractor


class GLC_Gas(TriomaClass):
    """
    GLC_Gas class represents a sweep gas in a GLC (Gas-Liquid Contactor) system.
    Attributes:
        G_gas (float): The flow rate of the gas.
        pg_in (float, optional): The Tritium inlet partial pressure of the gas. Default is 0.
        p_tot (float, optional): The total pressure of the component. Default is 100000 Pa.
    """

    def __init__(
        self,
        G_gas: float = None,
        pg_in: float = 0,
        pg_out: float = 0,
        p_tot: float = 100000,
    ):
        """
        Initializes a new instance of the GLC_Gas class.
        Args:
            G_gas (float): The flow rate of the gas.
            pg_in (float, optional): The Tritium inlet partial pressure of the gas. Default is 0.
            p_tot (float, optional): The total pressure of the component. Default is 100000 Pa.
            kla (float, optional): The total (kl*Area) mass transfer coefficient. Default is 0.
        """
        self.G_gas = G_gas
        self.pg_in = pg_in
        self.p_tot = p_tot
        self.pg_out = pg_out


class GLC(TriomaClass):
    """
    GLC (Gas-Liquid Contact) class represents a gas-liquid contactor component.
    Args:
        H (float): Height of the GLC.
        R (float): Radius of the GLC.
        L (float): Characteristic Length of the GLC fluid flow.
        c_in (float): Inlet concentration of the GLC.
        eff (float, optional): Efficiency of the GLC. Defaults to None.
        fluid (Fluid, optional): Fluid object representing the liquid phase. Defaults to None.
        membrane (Membrane, optional): Membrane object representing the membrane used in the GLC. Not super important. Defaults to None.
        GLC_gas (GLC_Gas, optional): GLC_Gas object representing the gas phase. Defaults to None.
    Attributes:
        H (float): Height of the GLC [m].
        R (float): Radius of the GLC [m].
        L (float): Length of the GLC [m].
        GLC_gas (GLC_Gas): GLC_Gas object representing the gas phase.
    Methods:
        get_kla_Ring(): Calculates the mass transfer coefficient (kla) for a Raschig Ring matrix.
    """

    def __init__(
        self,
        H: float = None,
        R: float = None,
        L: float = None,
        c_in: float = None,
        eff: float = None,
        fluid: "Fluid" = None,
        membrane: "Membrane" = None,
        GLC_gas: "GLC_Gas" = None,
        T: float = None,
        G_L: float = None,
        c_out: float = None,
        kla: float = None,
    ):
        """
        Initializes a new instance of the GLC class.
        Args:
            H (float): Height of the GLC.
            R (float): Radius of the GLC.
            L (float): Characteristic Length of the GLC fluid flow.
            c_in (float): Inlet concentration of the GLC.
            c_out (float): Outlet concentration of the GLC.
            eff (float, optional): Efficiency of the GLC. Defaults to None.
            fluid (Fluid, optional): Fluid object representing the liquid phase. Defaults to None.
            membrane (Membrane, optional): Membrane object representing the membrane used in the GLC. Not super important. Defaults to None.
            GLC_gas (GLC_Gas, optional): GLC_Gas object representing the gas phase. Defaults to None.
        """
        self.c_in = c_in
        self.eff = eff
        self.fluid = fluid
        self.H = H
        self.R = R
        self.L = L
        self.G_L = G_L
        self.T = T
        self.GLC_gas = GLC_gas
        self.c_out = c_out
        self.eff = eff
        self.kla = kla

    # def get_kla_Ring(self):
    #     """
    #     Calculates the mass transfer coefficient (kla) for a Raschig ring matrix.
    #     The mass transfer coefficient is calculated based on the Reynolds number (Re) and Schmidt number (Sc) of the fluid,
    #     as well as the diameter (d) of the ring.
    #     Returns:
    #         None
    #     """
    #     d = 2e-3  # Ring diameter
    #     Re = corr.Re(rho=self.fluid.rho, u=self.fluid.U0, L=self.L, mu=self.fluid.mu)
    #     Sc = corr.Schmidt(D=self.fluid.D, mu=self.fluid.mu, rho=self.fluid.rho)
    #     self.kla = extractor.corr_packed(
    #         Re,
    #         Sc,
    #         d,
    #         rho_L=self.fluid.rho,
    #         mu_L=self.fluid.mu,
    #         L=self.L,
    #         D=self.fluid.D,
    #     )

    def get_c_out(self):
        """
        Calculates the outlet concentration of the GLC.
        The outlet concentration is calculated based on the inlet concentration, the efficiency of the GLC, and the fluid properties.
        Returns:
            None
        """
        match self.fluid.MS:
            case False:
                c_out, eff = extractor.get_c_out_GLC_lm(
                    Z=self.H,
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in**2 / self.fluid.Solubility**2,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_S=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                    kla=self.kla,
                )
                self.eff = eff
                self.c_out = c_out
            case True:
                c_out, eff = extractor.get_c_out_GLC_ms(
                    Z=self.H,
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in / self.fluid.Solubility,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_H=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                    kla=self.kla,
                )
                self.eff = eff
                self.c_out = c_out

    def get_kla_from_cout(self):
        match self.fluid.MS:
            case False:
                Bl, kla = extractor.extractor_lm(
                    Z=self.H,
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in**2 / self.fluid.Solubility**2,
                    pl_out=self.c_out**2 / self.fluid.Solubility**2,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_S=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                )
                self.kla = kla
                self.Bl = Bl

            case True:
                Bl, kla = extractor.extractor_ms(
                    Z=self.H,
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in / self.fluid.Solubility,
                    pl_out=self.c_out / self.fluid.Solubility,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_H=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                )
                self.kla = kla
                self.Bl = Bl

        return Bl, kla

    def get_z_from_eff(self):
        """
        Calculates the height of the GLC from the efficiency of the GLC.
        The height is calculated based on the efficiency of the GLC and the radius of the GLC.
        Returns:
            None
        """
        match self.fluid.MS:
            case False:
                z = extractor.length_extractor_lm(
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in**2 / self.fluid.Solubility**2,
                    pl_out=self.c_out**2 / self.fluid.Solubility**2,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_S=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                    kla=self.kla,
                )

            case True:
                z = extractor.length_extractor_ms(
                    R=self.R,
                    G_l=self.G_L,
                    G_gas=self.GLC_gas.G_gas,
                    pl_in=self.c_in / self.fluid.Solubility,
                    pl_out=self.c_out / self.fluid.Solubility,
                    T=self.T,
                    p_t=self.GLC_gas.p_tot,
                    K_H=self.fluid.Solubility,
                    pg_in=self.GLC_gas.pg_in,
                    kla=self.kla,
                )

        return z

    # def connect_to_component(
    #     self, component2: Union["Component", "BreedingBlanket"] = None
    # ):
    #     """sets the inlet conc of the object component equal to the outlet of self"""
    #     component2.update_attribute("c_in", self.c_out)

    def connect_to_component(self):
        return



from TRIOMA.tools.TriomaClass import TriomaClass
from TRIOMA.tools.Extractors.PipeSubclasses import (
    Geometry,
    Fluid,
    Membrane,
    FluidMaterial,
    SolidMaterial,
    Turbulator,
    WireCoil,
    CustomTurbulator,
)
from TRIOMA.tools.Extractors.GasLiquidContactor import GLC_Gas, GLC

import TRIOMA.tools.correlations as corr
import matplotlib.pyplot as plt

from typing import Union
import matplotlib.pyplot as plt
from scipy import integrate
from TRIOMA.tools.Extractors import extractor
from TRIOMA.tools.Extractors.PAV import Component
from TRIOMA.tools.BreedingBlanket import BreedingBlanket
from TRIOMA.tools.Circuit import Circuit
import numpy as np


def connect_to_component(
    self, component2: Union["Component", "BreedingBlanket", "GLC"] = None
):
    """sets the inlet conc of the object component equal to the outlet of self"""
    component2.update_attribute("c_in", self.c_out)


# Component.converge_split_HX = converge_split_HX
# Component.split_HX = split_HX
Component.connect_to_component = connect_to_component
BreedingBlanket.connect_to_component = connect_to_component
GLC.connect_to_component = connect_to_component



from TRIOMA.tools.component_tools import FluidMaterial, SolidMaterial
import numpy as np

N_A = 6.022e23
k_b = 8.617e-5
R_const = 8.314


# Define the fluid material
def Flibe(T):

    def density(T):
        return 2413 - 0.488 * T

    def viscosity(T):
        return 1.16e-4 * np.exp(3755 / T)

    def H_diff(T):
        R_const = 8.314
        return 9.3e-7 * np.exp(-42e3 / (R_const * T))

    def k():
        # thermal conductivity W/m/K
        return 1.1

    def cp():
        # soecific heat J/kg/K
        return 2386

    def k_H(T):
        # Henry's constant mol/m^3/Pa
        return 4.54e-4

    # def k_H_low(T):
    #     # Henry's constant mol/m^3/Pa Malinauskas
    #     R_const = 8.314
    #     return 2.1e-6 * np.exp(-29e3 / (R_const * T))

    # def k_H_high(T):
    #     # Henry's constant mol/m^3/Pa Calderoni
    #     R_const = 8.314
    #     return 7.9e-2 * np.exp(-35e3 / (R_const * T))

    # def k_H_mid(T):
    #     return 0.06 / T * np.exp(-(11 * (293 - 0.12 * T) / T))

    Flibe = FluidMaterial(
        T,
        D=H_diff(T),
        Solubility=k_H(T),
        MS=True,
        mu=viscosity(T),
        rho=density(T),
        k=k(),
        cp=cp(),
    )

    return Flibe


def Sodium(T):
    def density(T):
        return 219 + 275.32 * (1 - T / 2504.7) + 511.58 * (1 - T / 2503.7) ** 0.5

    def viscosity(T):
        return np.exp(-6.4406 - 0.3958 * np.log(T) + 556.835 / T)

    def H_diff(T):
        R_const = 8.314
        return 2e-5 * np.exp(-49053 / (R_const * T))

    def k():
        # thermal conductivity W/m/K
        return 1  # TODO

    def cp():
        # soecific heat J/kg/K
        return 1  # TODO

    def k_S(T):
        # Henry's constant mol/m^3/Pa
        return 10 ** (0.86 - 122 / T)

    Sodium = FluidMaterial(
        T,
        D=H_diff(T),
        Solubility=k_S(T),
        MS=False,
        mu=viscosity(T),
        rho=density(T),
        k=k(),
        cp=cp(),
    )
    return Sodium


def LiPb(T):
    def density(T):
        return 9659.8  # TODO

    def viscosity(T):
        return 1  # TODO

    def H_diff(T):
        return 1  # TODO

    def k():
        # thermal conductivity W/m/K
        return 1  # TODO

    def cp():
        # soecific heat J/kg/K
        return 1  # TODO

    def k_S(T):
        # Henry's constant mol/m^3/Pa
        MM_LiPb = 180  # TODO
        return 4.7e-7 * np.exp(-9e3 / (R_const * T) * density(T) / MM_LiPb)  # TODO

    LiPb = FluidMaterial(
        T,
        D=H_diff(T),
        Solubility=k_S(T),
        MS=False,
        mu=viscosity(T),
        rho=density(T),
        k=k(),
        cp=cp(),
    )
    return LiPb


def Steel(T):
    # Define the solid material

    def H_diff(T):
        D_met = 5.81e-7 * np.exp(-66.3e3 / (R_const * T))
        return D_met

    def K_S(T):
        return 1

    Steel = SolidMaterial(T=T, D=H_diff(T), K_S=K_S(T))

    return Steel



import sys
import os

sys.path.append(os.path.abspath("."))
from src.TRIOMA.tools.correlations import (
    get_deltaTML,
    Nu_SiederTate,
    Nu_Gnielinsky,
    Nu_DittusBoelter,
    f_Pethukov,
    f_Haaland,
    get_h_from_Nu,
    Schmidt,
    Re,
    Pr,
    get_k_from_Sh,
    Sherwood,
    Sherwood_HT_analogy,
    get_length_HX,
    Sherwood_bubbles,
)
import numpy as np


def test_get_deltaTML():
    T_in_hot = 100.0
    T_out_hot = 80.0
    T_in_cold = 40.0
    T_out_cold = 20.0

    expected_deltaTML = ((T_in_hot - T_out_cold) - (T_out_hot - T_in_cold)) / np.log(
        (T_in_hot - T_out_cold) / (T_out_hot - T_in_cold)
    )

    deltaTML = get_deltaTML(T_in_hot, T_out_hot, T_in_cold, T_out_cold)

    assert deltaTML == expected_deltaTML, "Test failed: Incorrect deltaTML value"


def test_Nu_SiederTate():
    Re = 10000.0
    Pr = 0.7
    mu_w = 0.001
    mu_c = 0.002

    expected_Nu = 0.027 * Re ** (4 / 5) * Pr ** (0.3) * (mu_c / mu_w) ** (0.14)

    Nu = Nu_SiederTate(Re, Pr, mu_w, mu_c)

    assert Nu == expected_Nu, "Test failed: Incorrect Nusselt number value"


def test_Nu_Gnielinsky():
    Re = 10000.0
    Pr = 0.7
    f = 0.02

    expected_Nu = (
        (f / 8) * (Re - 1000) * Pr / (1 + 12.7 * (f / 8) ** 0.5 * (Pr ** (2 / 3) - 1))
    )

    Nu = Nu_Gnielinsky(Re, Pr, f)

    assert Nu == expected_Nu, "Test failed: Incorrect Nusselt number value"


def test_Nu_DittusBoelter():
    Re = 10000.0
    Pr = 0.7

    expected_Nu = 0.023 * Re ** (4 / 5) * Pr ** (0.4)

    Nu = Nu_DittusBoelter(Re, Pr)

    assert Nu == expected_Nu, "Test failed: Incorrect Nusselt number value"


def test_f_Pethukov():
    Re = 10000.0
    Pr = 0.7

    expected_f = (0.79 * np.log(Re) - 1.64) ** (-2)

    f = f_Pethukov(Re, Pr)

    assert f == expected_f, "Test failed: Incorrect friction factor value"


def test_get_h_from_Nu():
    Nu = 100.0
    k = 0.5
    D = 0.1

    expected_h = Nu * k / D

    h = get_h_from_Nu(Nu, k, D)

    assert h == expected_h, "Test failed: Incorrect heat transfer coefficient value"


def test_f_Haaland():
    Re = 10000.0
    e_D = 0.001

    expected_f = (-1.8 * np.log10((e_D / 3.7) ** 1.11 + 6.9 / Re)) ** (-2)

    f = f_Haaland(Re, e_D)

    assert f == expected_f, "Test failed: Incorrect friction factor value"


def test_Schmidt():
    D = 0.1
    mu = 0.001
    rho = 1000.0

    expected_Sc = mu / (rho * D)

    Sc = Schmidt(D, mu, rho)

    assert Sc == expected_Sc, "Test failed: Incorrect Schmidt number value"


def test_Sherwood():
    Sc = 0.7
    Re = 10000.0

    expected_Sh = 0.0096 * Re**0.913 * Sc**0.346

    Sh = Sherwood(Sc, Re)

    assert Sh == expected_Sh, "Test failed: Incorrect Sherwood number value"


def test_Sherwood_HT_analogy():
    Re = 10000.0
    Sc = 0.7

    expected_Sh = 0.023 * Re**0.8 * Sc**0.4

    Sh = Sherwood_HT_analogy(Re, Sc)

    assert Sh == expected_Sh, "Test failed: Incorrect Sherwood number value"


def test_get_k_from_Sh():
    Sh = 100.0
    L = 1.0
    D = 0.1

    expected_k = Sh * D / L

    k = get_k_from_Sh(Sh, L, D)

    assert k == expected_k, "Test failed: Incorrect mass transfer coefficient value"


def test_Re():
    rho = 1000.0
    u = 2.0
    L = 0.5
    mu = 0.001

    expected_Re = rho * u * L / mu

    Re_val = Re(rho, u, L, mu)

    assert Re_val == expected_Re, "Test failed: Incorrect Reynolds number value"


def test_Pr():
    c_p = 1000.0
    mu = 0.001
    k = 0.5

    expected_Pr = mu * c_p / k

    Pr_val = Pr(c_p, mu, k)

    assert Pr_val == expected_Pr, "Test failed: Incorrect Prandtl number value"


def test_get_length_HX():
    deltaTML = 20.0
    d_hyd = 0.1
    U = 100.0
    Q = 5000.0

    expected_L = Q / (U * np.pi * d_hyd * deltaTML)

    L = get_length_HX(deltaTML, d_hyd, U, Q)

    assert L == expected_L, "Test failed: Incorrect length of the heat exchanger value"


def test_Sherwood_bubbles():
    Sc = 0.7
    Re = 10000.0

    expected_Sh = 0.089 * Re**0.69 * Sc**0.33

    Sh = Sherwood_bubbles(Sc, Re)

    assert Sh == expected_Sh, "Test failed: Incorrect Sherwood number value"



import numpy as np


def Nu_SiederTate(Re, Pr, mu_w, mu_c):
    """
    Calculate the Nusselt number using the Sieder-Tate correlation.

    Parameters:
    - Re (float): Reynolds number
    - Pr (float): Prandtl number
    - mu_w (float): viscosity of the wall fluid
    - mu_c (float): viscosity of the core fluid

    Returns:
    - Nu (float): Nusselt number
    """
    return 0.027 * Re ** (4 / 5) * Pr ** (0.3) * (mu_c / mu_w) ** (0.14)


def Nu_Gnielinsky(Re, Pr, f):
    """Calculate the Nusselt number using the Gnielinsky correlation.
    Parameters:
    - Re (float): Reynolds number
    - Pr (float): Prandtl number
    - f (float): friction factor
    Returns:
    - Nu (float): Nusselt number
    """
    num = (f / 8) * (Re - 1000) * Pr
    den = 1 + 12.7 * (f / 8) ** 0.5 * (Pr ** (2 / 3) - 1)
    return num / den


def Nu_DittusBoelter(Re, Pr):
    """Calculate the Nusselt number using the Dittus-Boelter correlation.
    Parameters:
    - Re (float): Reynolds number
    - Pr (float): Prandtl number
    Returns:
    - Nu (float): Nusselt number
    """
    return 0.023 * Re ** (4 / 5) * Pr ** (0.4)


def f_Pethukov(Re, Pr):
    """Calculate the friction factor using the Pethukov correlation.
    Parameters:
    - Re (float): Reynolds number
    - Pr (float): Prandtl number
    Returns:
    - f (float): friction factor
    """
    return (0.79 * np.log(Re) - 1.64) ** (-2)


def get_h_from_Nu(Nu, k, D):
    """Calculate the heat transfer coefficient from the Nusselt number.
    Parameters:
    - Nu (float): Nusselt number
    - k (float): thermal conductivity
    - D (float): diameter
    Returns:
    - h (float): heat transfer coefficient
    """
    return Nu * k / D


def f_Haaland(Re, e_D):
    """Calculate the friction factor using the Haaland correlation.
    Parameters:
    - Re (float): Reynolds number
    - e_D (float): relative roughness
    Returns:
    - f (float): friction factor
    """
    return (-1.8 * np.log10((e_D / 3.7) ** 1.11 + 6.9 / Re)) ** (-2)


def Schmidt(D, mu, rho):
    """Calculate the Schmidt number.
    Parameters:
    - D (float): diffusivity
    - mu (float): viscosity
    - rho (float): density
    Returns:
    - Sc (float): Schmidt number
    """
    return mu / (rho * D)


def Sherwood(Sc, Re):
    """Calculate the Sherwood number.
    Parameters:
    - Sc (float): Schmidt number
    - Re (float): Reynolds number
    Returns:
    - Sh (float): Sherwood number
    """

    return 0.0096 * Re**0.913 * Sc**0.346


def Sherwood_HT_analogy(Re, Sc):
    """Calculate the Sherwood number using the heat transfer analogy.
    Parameters:
    - Re (float): Reynolds number
    - Sc (float): Schmidt number
    Returns:
    - Sh (float): Sherwood number
    """
    return 0.023 * Re**0.8 * Sc**0.4


def get_k_from_Sh(Sh, L, D):
    """Calculate the mass transfer coefficient from the Sherwood number.
    Parameters:
    - Sh (float): Sherwood number
    - L (float): length
    - D (float): diameter
    Returns:
    - k (float): mass transfer coefficient
    """
    return Sh * D / L


def Re(rho, u, L, mu):
    """Calculate the Reynolds number.
    Parameters:
    - rho (float): density
    - u (float): velocity
    - L (float): length
    - mu (float): viscosity
    Returns:
    - Re (float): Reynolds number
    """
    return rho * u * L / mu


def Pr(c_p, mu, k):
    """Calculate the Prandtl number.
    Parameters:
    - c_p (float): specific heat capacity
    - mu (float): viscosity
    - k (float): thermal conductivity
    Returns:
    - Pr (float): Prandtl number
    """
    return mu * c_p / k


def Sherwood_bubbles(Sc, Re):
    """Calculate the Sherwood number for bubbles.
    Parameters:
    - Sc (float): Schmidt number
    - Re (float): Reynolds number
    Returns:
    - Sh (float): Sherwood number
    """
    return 0.089 * Re**0.69 * Sc**0.33  ## Humrickhouse MS report


def get_length_HX(deltaTML, d_hyd, U, Q):
    """Calculate the length of the heat exchanger.
    Parameters:
    - deltaTML (float): log mean temperature difference
    - d_hyd (float): hydraulic diameter
    - U (float): overall heat transfer coefficient
    - Q (float): heat transfer rate
    Returns:
    - L (float): length of the heat exchanger
    """
    L = Q / (U * np.pi * d_hyd * deltaTML)
    return L


def get_deltaTML(T_in_hot, T_out_hot, T_in_cold, T_out_cold):
    """Calculate the log mean temperature difference.
    Parameters:
    - T_in_hot (float): hot inlet temperature
    - T_out_hot (float): hot outlet temperature
    - T_in_cold (float): cold inlet temperature
    - T_out_cold (float): cold outlet temperature
    Returns:
    - deltaTML (float): log mean temperature difference
    """
    delta_T1 = T_in_hot - T_out_cold
    delta_T2 = T_out_hot - T_in_cold
    delta_T_lm = (delta_T1 - delta_T2) / np.log(delta_T1 / delta_T2)
    return delta_T_lm



import numpy as np
from typing import Union

from TRIOMA.tools import correlations as corr
from TRIOMA.tools.TriomaClass import TriomaClass


class Geometry(TriomaClass):
    """
    Represents the geometry of a component.

    Args:
        L (float): Length of the component.
        D (float): Diameter of the component.
        thick (float): Thickness of the component.
        n_pipes (int, optional): The number of pipes in the component. Defaults to 1.

    """

    def __init__(
        self,
        L: float = None,
        D: float = None,
        thick: float = None,
        n_pipes: float = 1,
        turbulator: Union["Turbulator"] = None,
    ):
        self.L = L
        self.D = D
        self.thick = thick
        self.n_pipes = n_pipes
        self.turbulator = turbulator

    def get_fluid_volume(self):
        """
        Calculates the volume of the fluid component.
        """

        return np.pi * (self.D / 2) ** 2 * self.L

    def get_solid_volume(self):
        """
        Calculates the volume of the solid component.
        """
        return np.pi * ((self.D / 2) ** 2 - (self.D / 2 - self.thick) ** 2) * self.L

    def get_total_volume(self):
        """
        Calculates the total volume of the component.
        """
        return self.get_fluid_volume() + self.get_solid_volume()


class Fluid(TriomaClass):
    """
    Represents a fluid in a component for Tritium transport analysis

    Args:
        T (float): Temperature of the fluid.
        D (float): Tritium Diffusivity of the fluid.
        D_0 (float): Preexponential Diffusivity of the fluid.
        E_d (float): Activation energy of the fluid diffusivity.
        Solubility (float): Solubility of the fluid.
        Solubility_0 (float): Preexponential solubility of the fluid.
        E_s (float): Activation energy of the fluid solubility.
        MS (bool): Indicates whether the fluid is a molten salt or a liquid metal.
        d_Hyd (float, optional): Hydraulic diameter of the fluid. Defaults to None.
        k_t (float, optional): Mass transport coefficient of the fluid. Defaults to None.
        mu (float, optional): Viscosity of the fluid. Defaults to None.
        rho (float, optional): Density of the fluid. Defaults to None.
        U0 (float, optional): Velocity of the fluid. Defaults to None.
        inv (float, optional): Inventory of the fluid. Defaults to None.
        recirculation(float,optional): fraction of recirculated flowrate. Defaults to 0. 1 = 100%, 0.5=50%.
    """

    def __init__(
        self,
        T: float = None,
        D: float = None,
        D_0: float = None,
        E_d: float = None,
        Solubility: float = None,
        Solubility_0: float = None,
        E_s: float = None,
        MS: bool = True,
        d_Hyd: float = None,
        k_t: float = None,
        mu: float = None,
        rho: float = None,
        U0: float = None,
        k: float = None,
        cp: float = None,
        inv: float = None,
        recirculation: float = 0,
        V: float = None,
    ):
        """
        Initializes a new instance of the Fluid class.

        Args:
            T (float): Temperature of the fluid.
            D (float): T Diffusivity of the fluid.
            Solubility (float): Solubility of the fluid.
            MS (bool): Indicates whether the fluid is a molten salt or a liquid metal.
            d_Hyd (float, optional): Hydraulic diameter of the fluid. Defaults to None.
            k_t (float, optional): Mass transport coefficient of the fluid. Defaults to None.
            mu (float, optional): Viscosity of the fluid. Defaults to None.
            rho (float, optional): Density of the fluid. Defaults to None.
            U0 (float, optional): Velocity of the fluid. Defaults to None.
            k thermal conductivity of the fluid. Defaults to None.
        """
        self.T = T
        self.MS = MS
        self.D_0 = D_0
        self.E_d = E_d
        self.Solubility_0 = Solubility_0
        self.E_s = E_s
        if D_0 is not None and E_d is not None:
            self.D = D_0 * np.exp(-E_d / (8.617333262145e-5 * self.T))
        else:
            self.D = D
        if Solubility_0 is not None and E_s is not None:
            self.Solubility = Solubility_0 * np.exp(-E_s / (8.617333262145e-5 * self.T))
        else:
            self.Solubility = Solubility
        self.k_t = k_t
        self.d_Hyd = d_Hyd
        self.mu = mu
        self.rho = rho
        self.recirculation = recirculation
        if self.recirculation == 0:
            self.U0 = U0
        else:
            self.U0 = U0 * (1 + self.recirculation)
        self.k = k
        self.cp = cp
        self.inv = inv

    def set_properties_from_fluid_material(
        self, fluid_material: "FluidMaterial" = None
    ):
        """
        Sets the properties of the fluid from a FluidMaterial object.

        Args:
            fluid_material (FluidMaterial): The FluidMaterial object to set the properties from.
        """
        self.T = fluid_material.T
        self.D = fluid_material.D
        self.Solubility = fluid_material.Solubility
        self.mu = fluid_material.mu
        self.rho = fluid_material.rho
        self.cp = fluid_material.cp
        self.k = fluid_material.k

    def update_T_prop(self):
        if self.D_0 is not None and self.E_d is not None:
            self.D = self.D_0 * np.exp(-self.E_d / (8.617333262145e-5 * self.T))
        if self.Solubility_0 is not None and self.E_s is not None:
            self.Solubility = self.Solubility_0 * np.exp(
                -self.E_s / (8.617333262145e-5 * self.T)
            )
        return

    def get_kt(self, turbulator=None):
        """
        Calculates the mass transport coefficient (k_t) for the fluid.

        If the hydraulic diameter (d_Hyd) is defined, the mass transport coefficient is calculated using correlations.
        Otherwise, an error message is printed.

        Returns:
            None
        """
        if self.d_Hyd is None:
            print("Hydraulic Diameter is not defined")
            return
        if self.k_t is None:
            Re = corr.Re(rho=self.rho, u=self.U0, L=self.d_Hyd, mu=self.mu)
            Sc = corr.Schmidt(D=self.D, mu=self.mu, rho=self.rho)
            if turbulator is None:

                # if Re < 1e4 and Re > 2030:
                #     Sh = 0.015 * Re**0.83 * Sc**0.42  ## Stempien Thesis pg 155-157 TODO implement different Re ranges
                if Re > 2030:
                    Sh = 0.0096 * Re**0.913 * Sc**0.346  ##Getthem paper
                    # Sh = 0.023 * Re**0.8 * Sc**0.33
                else:
                    print(str(Re) + " indicates laminar flow")
                    Sh = 3.66
                    # raise ValueError("Reynolds number is too low")
                self.k_t = corr.get_k_from_Sh(
                    Sh=Sh,
                    L=self.d_Hyd,
                    D=self.D,
                )
            else:
                match turbulator.turbulator_type:
                    case "WireCoil":

                        self.k_t = turbulator.k_t_correlation(
                            Re=Re, Sc=Sc, d_hyd=self.d_Hyd, D=self.D
                        )
                    case "TwistedTape":
                        raise NotImplementedError("Twisted Tape not implemented yet")
                    case "Custom":
                        self.k_t = turbulator.k_t_correlation(
                            Re=Re, Sc=Sc, d_hyd=self.d_Hyd, D=self.D
                        )

        else:
            print("k_t is already defined")


class Membrane(TriomaClass):
    """
    Represents a metallic membrane of a component for H transport.

    Attributes:
        T (float): Temperature of the membrane.
        D (float): Diffusion coefficient of the membrane.
        thick (float): Thickness of the membrane.
        K_S (float): Solubility coefficient of the membrane.
        k_d (float, optional): Dissociation rate constant of the membrane. Defaults to None.
        k_r (float, optional): Recombination rate constant of the membrane. Defaults to None.
        k (float, optional): Thermal conductivity of the membrane. Defaults to None.
        D_0 (float, optional): Pre-exponential factor of the membrane. Defaults to None.Overwrites D if defined
        E_d (float, optional): Activation energy of the diffusivity in the membrane in eV. Defaults to None. Overwrites D if defined
        K_S_0 (float, optional): Pre-exponential factor of the solubility in the membrane. Defaults to None.Overwrites K_S if defined
        E_S (float, optional): Activation energy of the solubility in the membrane in eV. Defaults to None. Overwrites K_S if defined
        inv (float, optional): Inventory of the membrane in mol. Defaults to None.
    """

    def __init__(
        self,
        T: float = None,
        D: float = None,
        thick: float = None,
        K_S: float = None,
        k_d: float = None,
        k_r: float = None,
        k: float = None,
        D_0: float = None,
        E_d: float = None,
        K_S_0: float = None,
        E_S: float = None,
        inv: float = None,
        V: float = None,
    ):
        """
        Initializes a new instance of the Membrane class.

        Args:
            T (float): Temperature of the membrane.
            D (float): Diffusion coefficient of the membrane.
            thick (float): Thickness of the membrane.
            K_S (float): Solubility coefficient of the membrane.
            k_d (float, optional): Dissociation rate constant of the membrane. Defaults to 1e6.
            k_r (float, optional): Recombination rate constant of the membrane. Defaults to 1e6.
        """
        self.T = T
        if D_0 is not None and E_d is not None:
            self.D = D_0 * np.exp(-E_d / (8.617333262145e-5 * self.T))
        else:
            self.D = D
        if K_S_0 is not None and E_S is not None:
            self.K_S = K_S_0 * np.exp(-E_S / (8.617333262145e-5 * self.T))
        else:
            self.K_S = K_S
        self.thick = thick
        self.k_d = k_d
        self.k_r = k_r
        self.k = k
        self.D_0 = D_0
        self.E_d = E_d
        self.inv = inv
        self.K_S_0 = K_S_0
        self.E_S = E_S

    def set_properties_from_solid_material(
        self, solid_material: "SolidMaterial" = None
    ):
        """
        Sets the properties of the membrane from a SolidMaterial object.

        Args:
            solid_material (SolidMaterial): The SolidMaterial object to set the properties from.
        """
        self.T = solid_material.T
        self.D = solid_material.D
        self.K_S = solid_material.K_S

    def update_T_prop(self):
        if self.D_0 is not None and self.E_d is not None:
            self.D = self.D_0 * np.exp(-self.E_d / (8.617333262145e-5 * self.T))
        if self.K_S_0 is not None and self.E_S is not None:
            self.K_S = self.K_S_0 * np.exp(-self.E_S / (8.617333262145e-5 * self.T))
        return


class FluidMaterial(TriomaClass):
    """
    Represents a fluid material with various properties.

    Attributes:
        T (float): Temperature of the fluid material.
        D (float): Density of the fluid material.
        Solubility (float): Solubility of the fluid material.
        MS (float): Molecular weight of the fluid material.
        mu (float): Viscosity of the fluid material.
        rho (float): Density of the fluid material.
        k (float): Thermal conductivity of the fluid material.
        cp (float): Specific heat capacity of the fluid material.
    """

    def __init__(
        self,
        T: float = None,
        D: float = None,
        Solubility: float = None,
        MS: bool = None,
        mu: float = None,
        rho: float = None,
        k: float = None,
        cp: float = None,
    ):
        self.T = T
        self.D = D
        self.Solubility = Solubility
        self.MS = MS
        self.mu = mu
        self.rho = rho
        self.k = k
        self.cp = cp


class SolidMaterial(TriomaClass):
    """
    Represents a solid material used in a component.

    Attributes:
        D (float): The Diffusivity of the solid material.
        K_S (float): The Sievert constant of the solid material.
    """

    def __init__(
        self, T: float = None, D: float = None, K_S: float = None, k: float = None
    ):
        self.T = T
        self.D = D
        self.K_S = K_S
        self.k = k


class Turbulator(TriomaClass):
    """
    Represents a turbulator in a component for Tritium transport analysis

    Args:
        turbulator_type (str): Type of the turbulator.
        turbulator_params (dict): Parameters of the turbulator.
    """

    def __init__(
        self,
        turbulator_type: str = None,
    ):
        """
        Initializes a new instance of the Turbulator class.

        Args:
            turbulator_type (str): Type of the turbulator.
            turbulator_params (dict): Parameters of the turbulator.
        """
        self.turbulator_type = turbulator_type


class WireCoil(Turbulator):
    """
    Represents a wire coil turbulator in a component for Tritium transport analysis

    Args:
        turbulator_type (str): Type of the turbulator.
        turbulator_params (dict): Parameters of the turbulator.
    """

    def __init__(
        self,
        turbulator_type: str = "WireCoil",
        pitch: float = None,
    ):
        """
        Initializes a new instance of the WireCoil class.

        Args:
            turbulator_type (str): Type of the turbulator.
            pitch (float): Pitch of the wire coil.
        """
        super().__init__(turbulator_type)
        self.pitch = pitch

    def k_t_correlation(
        self, Re: float = None, Sc: float = None, d_hyd: float = None, D: float = None
    ):
        """
        Calculates the mass transport coefficient (k_t) for the fluid with a wire coil turbulator.

        Args:
            Re (float): Reynolds number of the fluid.
            Sc (float): Schmidt number of the fluid.

        Returns:
            float: The mass transport coefficient.
        """
        if Re > 2030:
            Sh = 0.132 * Re**0.72 * Sc**0.37 * (self.pitch / d_hyd) ** -0.372
        else:
            Sh = 3.66
        k_t = corr.get_k_from_Sh(Sh=Sh, L=self.pitch, D=D)
        return k_t

    def h_t_correlation(
        self, Re: float = None, Pr: float = None, d_hyd: float = None, k=None
    ):
        """
        Calculates the heat transfer coefficient (h_t) for the fluid with a wire coil turbulator.

        Args:
            Re (float): Reynolds number of the fluid.
            Pr (float): Prandtl number of the fluid.

        Returns:
            float: The heat transfer coefficient.
        """
        if Re > 2030:
            Nu = 0.132 * Re**0.72 * Pr**0.37 * (self.pitch / d_hyd) ** -0.372
        else:
            Nu = 3.66
        h_t = corr.get_h_from_Nu(Nu=Nu, k=k, D=d_hyd)
        return h_t


class CustomTurbulator(Turbulator):
    """
    Represents a custom turbulator in a component for Tritium transport analysis
    """

    def __init__(
        self,
        turbulator_type: str = "Custom",
        a: float = None,
        b: float = None,
        c: float = None,
    ):
        """
        Initializes a new instance of the CustomTurbulator class.

        Args:
            turbulator_type (str): Type of the turbulator.
            turbulator_params (dict): Parameters of the turbulator.
        """
        super().__init__(turbulator_type)
        self.a = a
        self.b = b
        self.c = c

    def k_t_correlation(
        self, Re: float = None, Sc: float = None, d_hyd: float = None, D: float = None
    ):
        """
        Calculates the mass transport coefficient (k_t) for the fluid with a custom turbulator.

        Args:
            Re (float): Reynolds number of the fluid.
            Sc (float): Schmidt number of the fluid.

        Returns:
            float: The mass transport coefficient.
        """
        if Re > 2030:
            Sh = self.a * Re**self.b * Sc**self.c
        else:
            Sh = 3.66
        k_t = corr.get_k_from_Sh(Sh=Sh, L=d_hyd, D=D)
        return k_t

    def h_t_correlation(
        self, Re: float = None, Pr: float = None, d_hyd: float = None, k: float = None
    ):
        """
        Calculates the heat transfer coefficient (h_t) for the fluid with a custom turbulator.

        Args:
            Re (float): Reynolds number of the fluid.
            Pr (float): Prandtl number of the fluid.

        Returns:
            float: The heat transfer coefficient.
        """
        if Re > 2030:
            Nu = self.a * Re**self.b * Pr**self.c
        else:
            Nu = 3.66
        h_t = corr.get_h_from_Nu(Nu=Nu, k=k, L=self.pitch)
        return h_t



import TRIOMA.tools.molten_salts as ms


def W(k_d, D, thick, K_S, c0, k_H):
    """
    Calculates the value of W using the given parameters.

    Parameters:
    k_d (float): The value of dissociation.
    D (float): The value of Diffusivity in the metal.
    thick (float): The value of thickness of the metal.
    K_S (float): The value of the sievert's constant.
    c0 (float): The value of the initial concentration of hydrogen in the molten salt.
    k_H (float): The value of Henri's constant.

    Returns:
    float: The calculated value of W.
    For W<<1, the regime is surface limited.
    """
    return 2 * k_d * thick * (c0 / k_H) ** 0.5 / (D * K_S)


def H(k_t, k_H, k_d):
    """
    Calculate the value of H dimensionless number based on the given parameters.

    Parameters:
    k_t (float): The value of mass transport coeff
    k_H (float): The value of Henri's constant.
    k_d (float): The value of dissociation.

    Returns:
    float: The calculated value of H.
    If H>>1, and H/W>>1 the regime is mass transport limited.
    """
    return k_d / (k_t * k_H)


def get_regime(k_d, D, thick, K_S, c0, k_t, k_H, print_var: bool = False):
    """
    Get the regime based on the value of H and W.
    k_d (float): The value of dissociation constant.
    D (float): The value of Diffusivity in the metal.
    thick (float): The value of thickness of the metal.
    K_S (float): The value of the sievert's constant in the metal.
    c0 (float): The value of the initial concentration of hydrogen in the molten salt.
    k_t (float): The value of mass transport coefficient.
    k_H (float): The value of Henri's constant.
    print_var (bool): If True, print the value of H, W and H/W.
    Returns:
    updates the values of Compontent.H and Component.W.
    if print_var is True, it will print the regime based on the value of H and W.
    str: The regime based on the value of H and W.
    """
    W = ms.W(k_d, D, thick, K_S, c0=c0, k_H=k_H)
    H = ms.H(k_t, k_H, k_d)
    result = None  ## initialize the result variable
    match (H, W):
        case (H, W) if H > 10 and H / W > 10:
            result = "Mass transport limited"
        case (H, W) if H < 0.1 and W < 0.1:
            result = "Surface limited"
        case (H, W) if W > 10 and H / W < 0.1:
            result = "Diffusion Limited"
        case _:
            result = "Mixed regime"
    if print_var:
        print("H is equal to", H)
        print("W is equal to", W)
        print("H/W is equal to", H / W)
        print(result)
    return result



import unittest
import sys
import os
import matplotlib

from TRIOMA.tools.Extractors.PipeSubclasses import CustomTurbulator

matplotlib.use("Agg")
sys.path.append(os.path.abspath("."))
import unittest
from src.TRIOMA.tools.component_tools import (
    Component,
    Fluid,
    Membrane,
    FluidMaterial,
    BreedingBlanket,
    SolidMaterial,
    Geometry,
    Circuit,
    WireCoil,
    GLC,
    GLC_Gas,
)
from src.TRIOMA.tools.materials import Flibe, Steel, Sodium, LiPb
from io import StringIO
from unittest.mock import patch
import pytest
import numpy as np
import matplotlib.pyplot as plt


class TestMSComponent(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some initial values
        self.T_fluid = 300
        self.D_fluid = 1e-9
        self.Solubility_fluid = 0.5
        self.MS_fluid = True
        self.mu_fluid = 1e-3
        self.rho_fluid = 1000
        self.k_fluid = 0.5
        self.cp_fluid = 1.0
        self.k_t_fluid = 0.1
        self.U0_fluid = 0.2
        self.d_Hyd_fluid = 0.3
        self.L_geom = 1.0
        self.thick_geom = 0.5
        self.D_geom = 0.3
        self.k_d_membrane = 1e7
        self.D_membrane = 0.4
        self.thick_membrane = 0.5
        self.K_S_membrane = 0.6
        self.T_membrane = 300
        self.k_r_membrane = 1e7
        self.k_membrane = 0.8
        self.c_in_component = 0.5
        self.eff_component = 0.8
        fluid = Fluid(
            T=300,
            D=1e-9,
            Solubility=0.5,
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=0.2,
            d_Hyd=0.3,
        )
        geometry = Geometry(L=1.0, thick=0.5, D=0.3)
        membrane = Membrane(D=0.4, thick=0.5, K_S=0.6, T=300, k_r=1e7, k=0.8, k_d=1e7)
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_outlet_c_comp(self):
        # Test the outlet_c_comp() method
        self.component.outlet_c_comp()
        self.assertAlmostEqual(
            self.component.c_out, self.c_in_component * (1 - self.eff_component)
        )

    def test_T_leak(self):
        # Test the T_leak() method
        leak = self.component.T_leak()
        self.assertAlmostEqual(
            leak,
            self.c_in_component
            * (self.eff_component)
            * self.D_geom**2
            / 4
            * np.pi
            * self.U0_fluid,
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mixed regime")

    def test_get_adimensionals(self):
        # Test the get_adimensionals() method
        self.component.get_adimensionals()
        self.assertEqual(
            self.component.H,
            self.component.membrane.k_d
            / (self.component.fluid.k_t * self.component.fluid.Solubility),
        )
        self.assertEqual(
            self.component.W,
            2
            * self.component.membrane.k_d
            * self.component.geometry.thick
            * (self.component.c_in / self.component.fluid.Solubility) ** 0.5
            / (self.component.membrane.D * self.component.membrane.K_S),
        )

    def test_use_analytical_efficiency(self):
        # Test the use_analytical_efficiency() method ##TODO: Hard coded test
        self.component.use_analytical_efficiency()
        self.assertAlmostEqual(self.component.eff, 0.99871670123992)

    def test_get_efficiency(self):
        # Test the get_efficiency() method
        self.component.get_efficiency()

        self.assertAlmostEqual(self.component.eff, 0.998984924629)

    def test_analytical_efficiency(self):
        # Test the analytical_efficiency() method
        self.component.analytical_efficiency()
        self.assertAlmostEqual(self.component.eff_an, 0.99871670123992)

    def test_get_flux(self):
        # Test the get_flux() method
        flux = self.component.get_flux(c=0.3, c_guess=0.3)
        self.assertAlmostEqual(flux, 0.0014967, places=5)

    def test_get_global_HX_coeff(self):
        # Test the get_global_HX_coeff() method
        U = self.component.get_global_HX_coeff(R_conv_sec=0.1)
        self.assertAlmostEqual(self.component.U, 2.9215784663)

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )

    def test_component_inventory(self):
        self.component.use_analytical_efficiency()
        self.component.get_inventory()
        self.assertEqual(self.component.inv, 0.00531677445914132)

    # def test_inspect(self):
    #     result = "c_in: 0.5\neff: 0.8\nL: 1.0\nfluid is a <class 'tools.component_tools.Fluid'> class, printing its variables:\n    T: 300\n    Solubility: 0.5\n    MS: True\n    D: 1e-09\n    k_t: 0.1\n    d_Hyd: 0.3\n    mu: 0.001\n    rho: 1000\n    U0: 0.2\n    k: 0.5\n    cp: 1.0\nmembrane is a <class 'tools.component_tools.Membrane'> class, printing its variables:\n    T: 300\n    D: 0.4\n    thick: 0.5\n    k_d: 10000000.0\n    K_S: 0.6\n    k_r: 10000000.0\n    k: 0.8\nH: None\nW: None"
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())
    #     result = ""
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect("kjhgfd")
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())


class TestLMComponent(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some initial values
        fluid = Fluid(
            T=300,
            D=1e-9,
            Solubility=0.5,
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=0.2,
            d_Hyd=0.3,
        )
        geometry = Geometry(L=1.0, thick=0.5, D=0.3)
        membrane = Membrane(k_d=1e7, D=0.4, thick=0.5, K_S=0.6, T=300, k_r=1e7, k=0.8)
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_outlet_c_comp(self):
        # Test the outlet_c_comp() method
        self.component.outlet_c_comp()
        self.assertAlmostEqual(
            self.component.c_out, self.component.c_in * (1 - self.component.eff)
        )

    def test_T_leak(self):
        # Test the T_leak() method
        leak = self.component.T_leak()
        self.assertAlmostEqual(
            leak,
            self.component.c_in
            * self.component.eff
            * self.component.geometry.D**2
            / 4
            * np.pi
            * self.component.fluid.U0,
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(regime, "Mixed regime")

    def test_get_adimensionals(self):
        # Test the get_adimensionals() method
        self.component.get_adimensionals()
        self.assertEqual(
            self.component.H,
            self.component.membrane.k_r
            / self.component.membrane.D
            * self.component.membrane.K_S
            * self.component.membrane.thick
            * self.component.c_in
            / self.component.fluid.Solubility
            * self.component.membrane.D
            * self.component.membrane.K_S
            / (
                self.component.fluid.k_t
                * self.component.fluid.Solubility
                * self.component.membrane.thick
            ),
        )
        self.assertEqual(
            self.component.W,
            self.component.membrane.k_r
            / self.component.membrane.D
            * self.component.membrane.K_S
            * self.component.membrane.thick
            * self.component.c_in
            / self.component.fluid.Solubility,
        )

    def test_use_analytical_efficiency(self):
        # Test the use_analytical_efficiency() method
        self.component.use_analytical_efficiency()
        self.assertAlmostEqual(self.component.eff, 0.998295638580)

    def test_get_efficiency(self):
        # Test the get_efficiency() method
        self.component.get_efficiency(c_guess=self.component.c_in / 2, plotvar=True)
        self.assertAlmostEqual(self.component.eff, 0.9986246, places=5)

    def test_analytical_efficiency(self):
        # Test the analytical_efficiency() method
        self.component.fluid.update_attribute("k_t", None)
        self.component.analytical_efficiency()
        self.assertAlmostEqual(self.component.eff_an, 0.00053628139636452)

    def test_get_flux(self):
        # Test the get_flux() method
        flux = self.component.get_flux(c=0.3)
        self.assertAlmostEqual(flux, 0.01314458525095)

    def test_get_global_HX_coeff(self):
        # Test the get_global_HX_coeff() method
        U = self.component.get_global_HX_coeff(R_conv_sec=0.1)
        self.assertAlmostEqual(self.component.U, 2.9215784663)

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )

    # def test_inspect(self):
    #     result = "c_in: 0.5\neff: 0.8\nL: 1.0\nfluid is a <class 'tools.component_tools.Fluid'> class, printing its variables:\n    T: 300\n    Solubility: 0.5\n    MS: False\n    D: 1e-09\n    k_t: 0.1\n    d_Hyd: 0.3\n    mu: 0.001\n    rho: 1000\n    U0: 0.2\n    k: 0.5\n    cp: 1.0\nmembrane is a <class 'tools.component_tools.Membrane'> class, printing its variables:\n    T: 300\n    D: 0.4\n    thick: 0.5\n    k_d: 10000000.0\n    K_S: 0.6\n    k_r: 10000000.0\n    k: 0.8\nH: None\nW: None"
    #     expected_output = result
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), expected_output.strip())

    def test_set_attribute(self):
        self.component.update_attribute("c_in", 0.6)
        self.assertEqual(self.component.c_in, 0.6)
        ##test internal attribute
        self.component.update_attribute("U0", 0.3)
        self.assertEqual(self.component.fluid.U0, 0.3)
        with pytest.raises(ValueError) as excinfo:
            self.component.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of Component"

    def test_set_material_properties(self):
        T = 800
        fluid_material = Flibe(T)
        self.component.fluid.set_properties_from_fluid_material(fluid_material)
        self.assertEqual(self.component.fluid.rho, fluid_material.rho)
        solid_material = Steel(T)
        self.component.membrane.set_properties_from_solid_material(solid_material)
        self.assertEqual(self.component.membrane.K_S, solid_material.K_S)

    def test_LiPb(self):
        T = 800
        fluid_material = LiPb(T)
        self.component.fluid.set_properties_from_fluid_material(fluid_material)
        self.assertEqual(self.component.fluid.rho, fluid_material.rho)
        solid_material = Steel(T)
        self.component.membrane.set_properties_from_solid_material(solid_material)
        self.assertEqual(self.component.membrane.K_S, solid_material.K_S)

    def test_sodiium(self):
        T = 800
        fluid_material = Sodium(T)
        self.component.fluid.set_properties_from_fluid_material(fluid_material)
        self.assertEqual(self.component.fluid.rho, fluid_material.rho)
        solid_material = Steel(T)
        self.component.membrane.set_properties_from_solid_material(solid_material)
        self.assertEqual(self.component.membrane.K_S, solid_material.K_S)

    def test_update_T_properties(self):
        T = 999
        self.component.update_attribute("T", T)
        self.assertEqual(self.component.fluid.T, T)

    def test_component_inventory(self):
        self.component.get_inventory()
        self.assertEqual(self.component.inv, 0.007008033266553368)
        inv_m_an = self.component.analytical_solid_inventory()
        inv_f_an = self.component.analytical_fluid_inventory()
        inv_m_num = self.component.get_solid_inventory(flag_an=False)
        inv_f_num = self.component.get_fluid_inventory(flag_an=False)
        self.assertAlmostEqual(inv_m_an, inv_m_num)
        self.assertAlmostEqual(inv_f_an, inv_f_num)

    def test_run_inspect(self):
        self.component.inspect()
        self.component.geometry.inspect()

    def test_pumping_power(self):
        self.component.get_pumping_power()
        self.assertAlmostEqual(self.component.pumping_power, 0.019029194163191265)

    def test_plot_component(self):
        self.component.use_analytical_efficiency()
        self.component.outlet_c_comp()
        self.component.plot_component()
        plt.close()

    def test_component_inventory(self):
        self.component.get_inventory(flag_an=False)
        self.assertAlmostEqual(self.component.inv, 0.0070080332665533665, places=7)


class TestExoticComponent(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some initial values
        self.T_fluid = 300
        self.D_fluid = 1e-9
        self.Solubility_fluid = 0.5
        self.MS_fluid = True
        self.mu_fluid = 1e-3
        self.rho_fluid = 1000
        self.k_fluid = 0.5
        self.cp_fluid = 1.0
        self.k_t_fluid = 0.1
        self.U0_fluid = 0.2
        self.d_Hyd_fluid = 0.3
        self.L_geom = 1.0
        self.thick_geom = 0.5
        self.D_geom = 0.3
        self.k_d_membrane = 1e7
        self.D_membrane = 0.4
        self.thick_membrane = 0.5
        self.K_S_membrane = 0.6
        self.T_membrane = 300
        self.k_r_membrane = 1e7
        self.k_membrane = 0.8
        self.c_in_component = 0.5
        self.eff_component = 0.8
        fluid = Fluid(
            T=300,
            D_0=1e-9,
            E_d=0.5,
            Solubility_0=0.5,
            E_s=0.5,
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=0.2,
            d_Hyd=0.3,
            recirculation=-0.5,
        )
        wirecoil = WireCoil(pitch=1e-2)
        geometry = Geometry(L=1.0, thick=0.5e-3, D=0.3, turbulator=wirecoil)
        membrane = Membrane(
            D_0=0.4,
            E_d=0.1,
            E_S=0.1,
            thick=0.5,
            K_S_0=0.6,
            T=300,
            k_r=1e7,
            k=0.8,
            k_d=1e7,
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_updateTproperties(self):
        self.assertEqual(self.component.fluid.D, 3.984462016634033e-18)
        self.assertEqual(self.component.membrane.D, 0.008358607447235438)
        T = 999
        self.component.update_attribute("T", T)
        self.assertEqual(self.component.fluid.T, T)
        self.assertAlmostEqual(self.component.membrane.D, 0.12519232082575535, places=7)
        self.assertEqual(self.component.fluid.D, 3.003229324075595e-12)

    def test_update_recirculation(self):
        self.assertAlmostEqual(self.component.fluid.recirculation, -0.5, places=5)
        self.assertAlmostEqual(self.component.fluid.U0, 0.1, places=5)

    def test_wirecoil_kt(self):
        self.component.fluid.update_attribute("k_t", None)
        self.component.use_analytical_efficiency()
        self.assertAlmostEqual(self.component.fluid.k_t, 3.74123654043251e-11)
        newturb = CustomTurbulator(a=1, b=1, c=1)
        self.component.geometry.turbulator = newturb
        self.component.fluid.update_attribute("k_t", None)
        self.component.use_analytical_efficiency()
        self.assertAlmostEqual(self.component.fluid.k_t, 0.1)

    def test_cout(self):
        self.component.outlet_c_comp()
        self.assertAlmostEqual(self.component.c_out, 0.3)
        self.component.fluid.update_attribute("recirculation", 0.5)
        self.component.outlet_c_comp()
        self.assertAlmostEqual(self.component.c_out, 5.399292025261743e-07)

    def test_flowrate(self):
        self.component.get_total_flowrate()
        self.assertAlmostEqual(self.component.flowrate, 0.007068583470577035)

    def test_volume(self):
        self.component.define_component_volumes()
        self.assertAlmostEqual(self.component.V, 0.07115628820564542)
        self.assertAlmostEqual(self.component.membrane.V, 0.0004704534998750733)
        self.assertAlmostEqual(self.component.fluid.V, 0.07068583470577035)
        self.component.geometry.get_total_volume()


class TestFluidMaterial(unittest.TestCase):
    def setUp(self):
        self.fluid_material = FluidMaterial(
            T=300, D=1e-9, Solubility=0.5, MS=True, mu=1e-3, rho=1000, k=0.5, cp=1.0
        )

    def test_attributes(self):
        self.assertEqual(self.fluid_material.T, 300)
        self.assertEqual(self.fluid_material.D, 1e-9)
        self.assertEqual(self.fluid_material.Solubility, 0.5)
        self.assertEqual(self.fluid_material.MS, True)
        self.assertEqual(self.fluid_material.mu, 1e-3)
        self.assertEqual(self.fluid_material.rho, 1000)
        self.assertEqual(self.fluid_material.k, 0.5)
        self.assertEqual(self.fluid_material.cp, 1.0)

    # def test_inspect(self):
    #     result = "T: 300\nD: 1e-09\nSolubility: 0.5\nMS: True\nmu: 0.001\nrho: 1000\nk: 0.5\ncp: 1.0"
    #     expected_output = result
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.fluid_material.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), expected_output.strip())

    def test_update_attribute(self):
        self.fluid_material.update_attribute("T", 400)
        self.assertEqual(self.fluid_material.T, 400)
        with pytest.raises(ValueError) as excinfo:
            self.fluid_material.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of FluidMaterial"


class Test_SolidMaterial(unittest.TestCase):
    def setUp(self):
        self.solid_material = SolidMaterial(T=300, K_S=1, D=1e-7)

    def test_attributes(self):
        self.assertEqual(self.solid_material.T, 300)
        self.assertEqual(self.solid_material.K_S, 1)

    # def test_inspect(self):
    #     result = "T: 300\nD: 1e-07\nK_S: 1\nk: None"
    #     expected_output = result
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.solid_material.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), expected_output.strip())

    def test_update_attribute(self):
        self.solid_material.update_attribute("T", 400)
        self.assertEqual(self.solid_material.T, 400)
        with pytest.raises(ValueError) as excinfo:
            self.solid_material.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of SolidMaterial"


class Test_BB_Component(unittest.TestCase):
    def setUp(self):
        self.component = BreedingBlanket(
            c_in=0, Q=0.5e9, TBR=1.05, T_out=900, T_in=800, fluid=Flibe(850)
        )

    def plot_test(self):
        fig = self.component.plot_component()

    def test_outlet_c_comp(self):
        # Test the outlet_c_comp() method
        self.component.get_cout(print_var=True)
        self.assertAlmostEqual(self.component.c_out, 0.0001473990666223908)

    def test_get_flowrate(self):
        self.component.get_flowrate()
        self.assertAlmostEqual(self.component.m_coolant, 2095.5574182607)

    def test_set_attribute(self):
        self.component.update_attribute("c_in", 0.6)
        self.assertEqual(self.component.c_in, 0.6)
        ##test internal attribute
        self.component.update_attribute("cp", 0.3)
        self.assertEqual(self.component.fluid.cp, 0.3)
        with pytest.raises(ValueError) as excinfo:
            self.component.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of BreedingBlanket"

    def test_plot_component(self):
        self.component.get_cout()
        self.component.plot_component()
        plt.close()

    # def test_inspect(self):
    #     result = "\n        c_in: 0\nQ: 500000000.0\nTBR: 1.05\nT_out: 900\nT_in: 800\nfluid is a <class 'tools.component_tools.FluidMaterial'> class, printing its variables:\n    T: 850\n    D: 2.4399672021371085e-09\n    Solubility: 0.000454\n    MS: True\n    mu: 0.009616515365587481\n    rho: 1998.2\n    k: 1.1\n    cp: 2386"
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())
    #     result = ""
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect("kjhgfd")
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())


class TestMembrane(unittest.TestCase):
    def setUp(self):
        self.component = Membrane(
            k_d=1e7, D=0.4, thick=0.5, K_S=0.6, T=300, k_r=1e7, k=0.8
        )

    # def test_inspect(self):
    #     result = "T: 300\nD: 0.4\nthick: 0.5\nk_d: 10000000.0\nK_S: 0.6\nk_r: 10000000.0\nk: 0.8"
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())
    #     result = ""
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect("kjhgfd")
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())

    def test_set_attribute(self):
        self.component.update_attribute("thick", 1e-3)
        self.assertEqual(self.component.thick, 1e-3)
        with pytest.raises(ValueError) as excinfo:
            self.component.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of Membrane"


class TestFluid(unittest.TestCase):
    def setUp(self):
        self.component = Fluid(
            T=300,
            D=1e-9,
            Solubility=0.5,
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=0.2,
            d_Hyd=0.3,
        )

    # def test_inspect(self):
    #     result = "T: 300\nSolubility: 0.5\nMS: True\nD: 1e-09\nk_t: 0.1\nd_Hyd: 0.3\nmu: 0.001\nrho: 1000\nU0: 0.2\nk: 0.5\ncp: 1.0"
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect()
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())
    #     result = ""
    #     with patch("sys.stdout", new=StringIO()) as fake_out:
    #         self.component.inspect("kjhgfd")
    #         self.assertEqual(fake_out.getvalue().strip(), result.strip())

    def test_set_attribute(self):
        self.component.update_attribute("Solubility", 1e-3)
        self.assertEqual(self.component.Solubility, 1e-3)
        with pytest.raises(ValueError) as excinfo:
            self.component.update_attribute("kghufh", 0.3)
        assert str(excinfo.value) == "'kghufh' is not an attribute of Fluid"

    def test_get_kt(self):
        # Test no d_hyd defined
        self.component.update_attribute("d_Hyd", None)
        result = "Hydraulic Diameter is not defined"
        with patch("sys.stdout", new=StringIO()) as fake_out:
            self.component.get_kt()
            self.assertEqual(fake_out.getvalue().strip(), result.strip())
        # test get_kt
        self.component.update_attribute("d_Hyd", 0.3)
        self.component.update_attribute("k_t", None)
        self.component.get_kt()
        self.assertEqual(self.component.k_t, 8.046408367835323e-06)
        # test when k_t is already defined
        # result = "k_t is already defined"
        # with patch("sys.stdout", new=StringIO()) as fake_out:
        #     self.component.get_kt()
        #     self.assertEqual(fake_out.getvalue().strip(), result.strip())
        # test when Reynolds number is too low
        self.component.update_attribute("U0", 1e-6)
        self.component.update_attribute("k_t", None)

        self.component.get_kt()
        self.assertEqual(
            (3.66 / self.component.d_Hyd * self.component.D), self.component.k_t
        )
        self.component.update_attribute("U0", 0.008)
        # test when Reynolds number is in a different correlation range
        self.component.get_kt()
        self.assertAlmostEqual(self.component.k_t, 1.2200000000000002e-08)


class TestMSComponentDiffusionLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e7, D=1e-9, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Diffusion Limited")

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestMSComponentMixedDiffusionMassTransport(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e7, D=1e-6, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mixed regime")

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestMSComponentMixedDiffusionSurface(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-2,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e-10, D=1e-7, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e-10, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mixed regime")

    def test_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in)


class TestMSComponentMixedMassTransportSurface(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-8,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e-8, D=1e-2, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e-8, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mixed regime")

    def test_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestMSComponentMassTransportLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e7, D=1e-2, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mass transport limited")

    def test_get_flux(self):
        # Test the get_flux() method
        flux = self.component.get_flux(c=0.3, c_guess=0.3)
        self.assertAlmostEqual(self.component.J_perm, -9.6149466095734e-05, places=5)

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestMSComponentSurfaceLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e-16, D=1e-9, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e-16, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Surface limited")

    def test_get_efficiency(self):
        # Test that get efficiency works
        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestMSComponentFullyMixed(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-11,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e-10, D=1e-7, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e-10, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime()
        self.assertEqual(regime, "Mixed regime")

    def test_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestLMComponentMassTransportLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e7, D=1e-2, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(
            regime,
            "Mass transport limited",
        )

    def test_get_flux(self):
        # Test the get_flux() method
        flux = self.component.get_flux(c=0.3, c_guess=0.3)
        self.assertAlmostEqual(self.component.J_perm, -4.80747330478670e-05, places=5)

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestLMComponentMixedDiffusionMassTransport(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-9,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e7, D=1e-7, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(
            regime,
            "Mixed regime",
        )

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestLMComponentMixedDiffusionSurface(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-3,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e-3, D=1e-7, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e-3, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(
            regime,
            "Mixed regime",
        )

    def test_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestLMComponentMixedMassTransferSurface(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-6,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e-3, D=1e-2, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e-3, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(
            regime,
            "Transport and surface limited regime",
        )

    def test_get_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestLMComponentFullyMixed(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-11,  # Changed diffusion coefficient
            Solubility=1e-2,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=2,
            d_Hyd=2e-3,
        )
        geometry = Geometry(L=1.0, thick=1e-4, D=2e-3)
        membrane = Membrane(
            k_d=1e-9, D=1e-9, thick=1e-4, K_S=0.6e-2, T=700, k_r=1e-9, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(
            regime,
            "Mixed regime",
        )

    def test_efficiency(self):
        # Test the efficiency_vs_analytical() method

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class TestLMComponentDiffusionLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-5,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e7, D=1e-9, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e7, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def plot_test(self):
        fig = self.component.plot_component()

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(regime, "Diffusion Limited")

    def test_get_flux(self):
        # Test the get_flux() method
        flux = self.component.get_flux(c=0.3)
        self.assertAlmostEqual(flux, 0.29990976788036605, places=5)

    def test_efficiency_vs_analytical(self):
        # Test the efficiency_vs_analytical() method
        self.component.analytical_efficiency()
        self.component.get_efficiency(c_guess=self.component.c_in / 2)
        self.assertAlmostEqual(
            abs(self.component.eff - self.component.eff_an) / self.component.eff_an,
            0,
            places=2,
        )


class TestLMComponentSurfaceLimited(unittest.TestCase):
    def setUp(self):
        # Create a Component object with some different initial values
        fluid = Fluid(
            T=750,  # Changed temperature
            D=2e-5,  # Changed diffusion coefficient
            Solubility=1e-4,  # Changed solubility
            MS=False,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=1.0, thick=1e-2, D=2e-2)
        membrane = Membrane(
            k_d=1e-16, D=1e-9, thick=1e-2, K_S=0.6e-2, T=700, k_r=1e-16, k=0.8
        )
        self.component = Component(
            c_in=0.5, geometry=geometry, eff=0.8, fluid=fluid, membrane=membrane
        )

    def plot_test(self):
        self.component.plot()

    def test_get_regime(self):
        regime = self.component.get_regime(print_var=True)
        self.assertEqual(regime, "Surface limited")

    def test_get_efficiency(self):
        # Test that get efficiency works

        self.component.get_efficiency(c_guess=self.component.c_in / 2)


class testclosedCircuit(unittest.TestCase):
    def setUp(self):
        fluid = Fluid(
            T=900,
            D=1e-7,
            Solubility=0.5,
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=1,
            d_Hyd=2e-2,
        )
        geometry = Geometry(L=10.0, thick=0.5e-3, D=2e-2)
        membrane = Membrane(
            D_0=1e-7, E_d=1, thick=0.5e-3, K_S=0.6, T=900, k_r=1e7, k=0.8, k_d=1e7
        )
        component = Component(
            geometry=geometry, fluid=fluid, membrane=membrane, loss=False
        )
        component2 = Component(
            geometry=geometry, fluid=fluid, membrane=membrane, loss=False, name="PAV"
        )
        componentBB = BreedingBlanket(
            c_in=1e-3,
            Q=0.5e9,
            TBR=1.05,
            T_out=900,
            T_in=800,
            fluid=Flibe(850),
            name="BB",
        )
        geometry_HX = Geometry(L=10.0, thick=1e-3, D=2e-3)
        membrane_HX = Membrane(
            D_0=1e-9, E_d=0.2, thick=1e-3, K_S=0.6, T=850, k_r=1e7, k=0.8, k_d=1e7
        )
        fluid_HX = Fluid(
            T=850,
            D=1e-9,
            Solubility=0.5,
            MS=True,
            mu=1e-3,
            rho=1000,
            k=0.5,
            cp=1.0,
            k_t=0.1,
            U0=1,
            d_Hyd=2e-2,
        )
        self.component_HX = Component(
            c_in=1e-3,
            geometry=geometry_HX,
            fluid=fluid_HX,
            membrane=membrane_HX,
            loss=True,
        )
        HX = self.component_HX.split_HX(
            N=11,
            T_in_hot=900,
            T_out_hot=800,
            T_in_cold=581,
            T_out_cold=800,
            R_sec=0,
            Q=0.5e9 / 5e3,
            plotvar=True,
        )
        plt.close()
        self.circuit = Circuit([componentBB, component, HX, component2], closed=True)

    def test_circuit(self):
        # Only test that Circuit functions are called and return correctly
        # Test closed circuit
        self.circuit.solve_circuit()
        self.circuit.get_eff_circuit()
        self.circuit.get_gains_and_losses()
        self.circuit.inspect_circuit()
        # fig=self.circuit.plot_circuit()
        # Test component split and then test open circuit
        self.component_HX.converge_split_HX(
            T_in_hot=900,
            T_out_hot=800,
            T_in_cold=581,
            T_out_cold=800,
            R_sec=0,
            Q=0.5e9 / 5e3,
            plotvar=True,
        )
        Circuit_HX = self.component_HX.split_HX(
            N=11,
            T_in_hot=900,
            T_out_hot=800,
            T_in_cold=581,
            T_out_cold=800,
            R_sec=0,
            Q=0.5e9 / 5e3,
        )
        Circuit_HX.solve_circuit()
        Circuit_HX.inspect_circuit()
        Circuit_HX.add_component(self.component_HX)
        Circuit_HX.solve_circuit()
        Circuit_HX.plot_circuit()
        metal_cost = np.ones(len(Circuit_HX.components))
        fluid_cost = np.ones(len(Circuit_HX.components))
        Circuit_HX.estimate_cost(metal_cost=metal_cost, fluid_cost=fluid_cost)
        plt.close()
        self.circuit.plot_circuit()

        plt.close()
        Circuit_HX.get_inventory()
        self.circuit.inspect_circuit(name="PAV")
        self.circuit.get_circuit_pumping_power()


class testLMGLCComponent(unittest.TestCase):
    def setUp(self):
        R_const = 8.314
        T = 400 + 273.15
        c_in = 1e-2
        c_out = 9e-3
        Z = 0.6
        R = 0.3
        SweepGas = GLC_Gas(G_gas=3 * 1e-3 / 3600, pg_in=0, p_tot=1.5e5)
        Q_l = 71.0 * 1e-3 / 3600
        Flibe = Fluid(Solubility=1.33e-4 * np.exp(-1350 / R_const / T), MS=False)
        Melodie = GLC(
            H=Z,
            R=R,
            c_in=c_in,
            fluid=Flibe,
            GLC_gas=SweepGas,
            T=T,
            G_L=Q_l,
            c_out=c_out,
        )
        self.GLC = Melodie

    def test_get_kla(self):
        self.GLC.get_kla_from_cout()
        self.assertAlmostEqual(self.GLC.kla, 1.2850594291115214e-05, places=8)

    def test_get_cout(self):
        self.GLC.kla = 1.2850594291115214e-05
        self.GLC.get_c_out()

        self.assertAlmostEqual(self.GLC.c_out, 0.009, places=7)

    def test_get_z_from_eff(self):
        self.GLC.kla = 1.2850594291115214e-05
        self.GLC.get_z_from_eff()
        self.assertAlmostEqual(self.GLC.H, 0.6, places=7)


class testMSGLCComponent(unittest.TestCase):
    def setUp(self):
        R_const = 8.314
        T = 400 + 273.15
        c_in = 1e-2
        c_out = 9e-3
        Z = 0.6
        R = 0.3
        SweepGas = GLC_Gas(G_gas=3 * 1e-3 / 3600, pg_in=0, p_tot=1.5e5)
        Q_l = 71.0 * 1e-3 / 3600
        Flibe = Fluid(Solubility=1.33e-4 * np.exp(-1350 / R_const / T), MS=True)
        Melodie = GLC(
            H=Z,
            R=R,
            c_in=c_in,
            fluid=Flibe,
            GLC_gas=SweepGas,
            T=T,
            G_L=Q_l,
            c_out=c_out,
        )
        self.GLC = Melodie

    def test_get_kla(self):
        self.GLC.get_kla_from_cout()
        self.assertAlmostEqual(self.GLC.kla, 2.9765872207306292e-05, places=8)

    def test_get_cout(self):
        self.GLC.kla = 2.9765872207306292e-05
        self.GLC.get_c_out()

        self.assertAlmostEqual(self.GLC.c_out, 0.009, places=7)

    def test_get_z_from_eff(self):
        self.GLC.kla = 2.9765872207306292e-05
        self.GLC.get_z_from_eff()
        self.assertAlmostEqual(self.GLC.H, 0.6, places=7)


if __name__ == "__main__":
    unittest.main()



import numpy
import TRIOMA.tools.correlations as cor
import scipy.integrate as integrate


def calculate_gas_velocity(G_gas, p_t, T, R):
    """__summary__
    Args:
        G_gas (float): Gas flowrate in normal conditions
        p_t (float): Total pressure
        T (float): Temperature
        R (float): Radius
    Returns:
        float: Gas velocity


    """
    Area = numpy.pi * R**2
    p_atm = 101325
    u_g = G_gas / Area / p_t * p_atm * T / 288.15
    return u_g


def extractor_lm(Z, R, G_l, G_gas, pl_in, pl_out, T, p_t, K_S, pg_in):
    """_summary_
    Args:
        Z (float): Height
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t pressure Pa of the column
        K_S Sievert's constant
    Returns:
        B_l liquid load
        k_la mass transfer coefficient in packed column
    """
    Area = numpy.pi * R**2  ## Area of the column
    B_l = (G_l) / Area * 3600  ## Liquid holdup
    u_l = G_l / Area  # Liquid velocity
    integral = NTU_lm(
        R, G_l, G_gas, pl_in, pl_out, T, p_t, K_S, pg_in
    )  ## number of transfer units,
    kla_c = u_l / Z * integral
    return [B_l, kla_c]


def length_extractor_lm(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_S, pg_in, kla, c_max=0):
    """_summary_
    Args:
        Z (float): Height
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t pressure Pa of the column
        K_S Sievert's constant
    Returns:
        B_l liquid load
        k_la mass transfer coefficient in packed column
    """
    Area = numpy.pi * R**2
    u_l = G_l / Area  # Liquid velocity
    integral = NTU_lm(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_S, pg_in, c_max)
    Z = u_l / kla * integral
    return Z


def NTU_lm(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_S, pg_in, c_max=0):
    """
    solving integral equation from (5) of "The engineering sizing of the packed desorption column of hydrogen
    # isotopes from Pb–17Li eutectic alloy. A rate based model using
    # experimental mass transfer coefficients from a Melodie loop""
    """
    Area = numpy.pi * R**2
    u_l = G_l / Area  # Liquid velocity

    R_const = 8.314
    u_g = calculate_gas_velocity(G_gas=G_gas, p_t=p_t, T=T, R=R)
    R_g = 2 * u_g / u_l  ## gas on liquid ratio
    c_in = pl_in**0.5 * K_S  # inlet concentration in liquid
    c_out = pl_out**0.5 * K_S  # outlet concentration in liquid
    c_in_gas = pg_in / R_const / T

    def toint(c):
        value = 1 / (
            c - K_S * ((R_const * T / R_g) * (c - c_out + c_in_gas * R_g)) ** 0.5
        )
        return value

    c_g_max = pl_in / R_const / T  # maximum concentration in gas
    c_out_max = max(
        c_in - R_g * (c_g_max - pg_in / R_const / T),  # maximum gas stripping
        c_max,  # given input from equation
        (pg_in) ** 0.5 * K_S,  ## if liquid is in equilibrium with gas at outlet
    )

    integral = integrate.quad(toint, c_out, c_in, points=c_out_max, maxp1=1e3)
    if integral[0] < 0:
        ## for debugging reasons as for now
        print("Warning: negative value of the integral", integral)
        print(
            "c out is "
            + str(c_out)
            + " c in is "
            + str(c_in)
            + " c in gas is "
            + str(c_in_gas)
            + " p in liquid is "
            + str(pl_in)
            + " p in gas is "
            + str(pg_in)
            + "solubility is "
            + str(K_S)
            + "rg is "
            + str(R_g)
        )
        print(" c out max is " + str(c_out_max))
    return integral[0]


def NTU_ms(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_H, pg_in, c_max=0):
    """
    solving integral equation from (5) of "The engineering sizing of the packed desorption column of hydrogen
    # isotopes from Pb–17Li eutectic alloy. A rate based model using
    # experimental mass transfer coefficients from a Melodie loop""
    # but for molten salts by changing the evolution of c star following the Henry's law
    """
    Area = numpy.pi * R**2
    u_l = G_l / Area  # Liquid velocity
    c_in = pl_in * K_H
    c_out = pl_out * K_H
    R_const = 8.314
    u_g = calculate_gas_velocity(G_gas=G_gas, p_t=p_t, T=T, R=R)
    c_in_gas = pg_in / R_const / T
    R_g = u_g / u_l  ## gas on liquid ratio

    def toint(c):
        return 1 / (
            c - K_H * (u_l / u_g * R_const * T) * (c - c_out + c_in_gas * u_g / u_l)
        )

    c_g_max = pl_in / R_const / T  # maximum concentration in gas
    c_out_max = max(
        c_in - R_g * (c_g_max - pg_in / R_const / T),  # maximum gas stripping
        c_max,  # given input from equation
        pg_in * K_H,  ## if liquid is in equilibrium with gas at outlet
    )

    integral = integrate.quad(toint, c_out, c_in, points=c_out_max, maxp1=1e3)
    integral = integrate.fixed_quad(toint, c_out, c_in)
    return integral[0]


def extractor_ms(Z, R, G_l, G_gas, pl_in, pl_out, T, p_t, K_H, pg_in):
    """_summary_
    Args:
        Z (float): Height
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t pressure Pa of the column
        K_S Sievert's constant
    Returns:
        B_l liquid load
        k_la mass transfer coefficient in packed column
    """
    Area = numpy.pi * R**2
    B_l = (G_l) / Area * 3600
    u_l = G_l / Area  # Liquid velocity
    integral = NTU_ms(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_H, pg_in)
    kla_c = u_l / Z * integral
    return [B_l, kla_c]


def length_extractor_ms(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_H, pg_in, kla, c_max=0):
    """_summary_
    Args:
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t(float): pressure Pa of the column
        K_S(float): Sievert's constant
    Returns:
        B_l (float):liquid load
        k_la(float): mass transfer coefficient in packed column
    """
    Area = numpy.pi * R**2
    u_l = G_l / Area  # Liquid velocity
    integral = NTU_ms(R, G_l, G_gas, pl_in, pl_out, T, p_t, K_H, pg_in, c_max=c_max)

    Z = u_l / kla * integral
    return Z


from scipy.optimize import minimize, root


def get_c_out_GLC_lm(Z, R, G_l, G_gas, pl_in, T, p_t, K_S, pg_in, kla):
    """_summary_
    Args:
        Z (float): Height
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t pressure Pa of the column
        K_S Sievert's constant
    Returns:
        B_l liquid load
        k_la mass transfer coefficient in packed column
    """
    u_l = G_l / (numpy.pi * R**2)
    c_in = pl_in**0.5 * K_S
    c_out_max_reaction = c_in - kla * Z / u_l * (
        c_in - pg_in**0.5 * K_S
    )  ## concentration at the outlet assuming maximum reaction rate possible
    R_const = 8.314
    Area = numpy.pi * R**2
    u_g = calculate_gas_velocity(G_gas=G_gas, p_t=p_t, T=T, R=R)
    c_g_max = pl_in / R_const / T  ## maximum gas concentration according with liquid
    c_out_max = max(
        c_in
        - u_g
        / u_l
        * 2
        * (
            c_g_max - pg_in / R_const / T
        ),  ## liquid concentration if gas strips as much as possible and gets into eq with liquid
        (pg_in) ** 0.5
        * K_S,  ## liquid concentration if it gets in equilibrium with gas
        c_out_max_reaction,  ## liquid concentration if reaction rate is at the maximum
    )

    def lenght_residual(c_out):
        pl_out_2 = c_out**2 / K_S**2
        z_guess = length_extractor_lm(
            R,
            G_l,
            G_gas,
            pl_in,
            pl_out_2,
            T,
            p_t,
            K_S,
            pg_in,
            kla,
            c_max=c_out_max,
        )
        return abs(float(Z - z_guess) ** 2)

    c_out = minimize(
        lenght_residual,
        c_out_max + (c_in - c_out_max) / 2,
        method="Powell",
        bounds=[(float(c_out_max), float(c_in))],
        tol=1e-20,
        options={"maxiter": 1e8, "xatol": 1e-20, "fatol": 1e-20},
    ).x[0]
    eff = 1 - c_out / c_in
    L_cout = length_extractor_lm(
        R, G_l, G_gas, pl_in, c_out**2 / K_S**2, T, p_t, K_S, pg_in, kla
    )
    if abs(L_cout - Z) > 1e-3:
        print(
            "Warning!: guessed length is not equal to the height. Double check your result",
            L_cout,
            Z,
        )
    if abs(c_out - c_out_max) < 1e-8:
        print("The sweep gas saturated")
        if L_cout < Z:
            print(" Longer column would not increment the extraction efficiency")
            eff = 1 - c_out / c_in
    return c_out, eff


def get_c_out_GLC_ms(Z, R, G_l, G_gas, pl_in, T, p_t, K_H, pg_in, kla):
    """_summary_
    Args:
        Z (float): Height
        R (float): Radius
        G_l (float): Liquid flowrate
        G_gas (float): Gas flowrate
        pl_in (float): T pressure inlet
        pl_out (float): T pressure outlet
        T (float): Temperature
        p_t pressure Pa of the column
        K_S Sievert's constant
    Returns:
        B_l liquid load
        k_la mass transfer coefficient in packed column
    """
    u_l = G_l / (numpy.pi * R**2)
    c_in = pl_in * K_H
    c_out_max_reaction = c_in - kla * Z / u_l * (
        c_in - pg_in * K_H
    )  ## concentration at the outlet assuming maximum reaction rate possible
    R_const = 8.314
    Area = numpy.pi * R**2
    u_g = calculate_gas_velocity(G_gas=G_gas, p_t=p_t, T=T, R=R)
    c_g_max = pl_in / R_const / T  ## maximum gas concentration according with liquid
    c_out_max = max(
        c_in
        - u_g
        / u_l
        * (
            c_g_max - pg_in / R_const / T
        ),  ## liquid concentration if gas strips as much as possible and gets into eq with liquid
        (pg_in) * K_H,  ## liquid concentration if it gets in equilibrium with gas
        c_out_max_reaction,  ## liquid concentration if reaction rate is at the maximum
    )

    def lenght_residual(c_out):
        pl_out_2 = c_out / K_H
        z_guess = length_extractor_ms(
            R,
            G_l,
            G_gas,
            pl_in,
            pl_out_2,
            T,
            p_t,
            K_H,
            pg_in,
            kla,
            c_max=c_out_max,
        )
        return abs(float(Z - z_guess) ** 2)

    c_out = minimize(
        lenght_residual,
        c_out_max + (c_in - c_out_max) / 2,
        method="Powell",
        bounds=[(float(c_out_max), float(c_in))],
        tol=1e-20,
        options={"maxiter": 1e8, "xatol": 1e-20, "fatol": 1e-20},
    ).x[0]
    eff = 1 - c_out / c_in
    L_cout = length_extractor_lm(
        R, G_l, G_gas, pl_in, c_out / K_H, T, p_t, K_H, pg_in, kla
    )
    if abs(L_cout - Z) > 1e-3:
        print(
            "Warning!: guessed length is not equal to the height. Double check your result",
            L_cout,
            Z,
        )
    if abs(c_out - c_out_max) < 1e-8:
        print("The sweep gas saturated")
        if L_cout < Z:
            print(" Longer column would not increment the extraction efficiency")
            eff = 1 - c_out / c_in
    return c_out, eff


def pack_corr(a, d, D, eta, v):
    k_l = (
        0.0051
        * (v / eta / a) ** (2 / 3)
        * (D / eta) ** 0.5
        * (a * d) ** 0.4
        * (1 / eta / 9.81) ** (-1 / 3)
    )
    return k_l


def corr_packed(Re, Sc, d, rho_L, mu_L, L, D):
    """_summary_
    Args
        Re (float): Reynolds
        Sc (float): Schmidt
        d (float): ring diameter
        rho_L (float): Liquid density
        mu_L (float): viscosity
        D diffusion coeff
        L= characteristic length
    Returns:
        float: _description_
        Warning: verification of this must be done
    """
    beta = 0.32  # Raschig rings 0.25
    g = 9.81
    Sh = beta * Re**0.59 * Sc**0.5 * (d**3 * g * rho_L**2 / mu_L**2) ** 0.17
    return cor.get_k_from_Sh(Sh, L, D)



from scipy.constants import physical_constants
from TRIOMA.tools.TriomaClass import TriomaClass
from scipy.constants import N_A
from TRIOMA.tools.Extractors.PipeSubclasses import Fluid
import matplotlib.pyplot as plt


class BreedingBlanket(TriomaClass):
    """
    Represents a breeding blanket component in a fuel cycle system.

    Attributes:
        c_in(float): Inlet concentration of tritium in the breeding blanket component.
        Q (float): Heat generated by the breeding blanket component.
        TBR (float): Tritium breeding ratio of the breeding blanket component.
        T_out (float): Outlet temperature of the breeding blanket component.
        T_in (float): Inlet temperature of the breeding blanket component.
        fluid (Fluid): Fluid used in the breeding blanket component.
    """

    def __init__(
        self,
        c_in: float = None,
        Q: float = None,
        TBR: float = None,
        T_out: float = None,
        T_in: float = None,
        fluid: Fluid = None,
        name: str = None,
        m_coolant: float = None,
    ):
        self.c_in = c_in
        self.Q = Q
        self.TBR = TBR
        self.T_out = T_out
        self.T_in = T_in
        self.fluid = fluid
        self.name = name
        self.m_coolant = m_coolant

    def plot_component(self):
        fig, ax2 = plt.subplots(1, 1, figsize=(10, 5))
        rectangle = plt.Rectangle(
            (0.2, 0.3), 0.55, 0.4, edgecolor="black", facecolor="green", alpha=0.5
        )
        ax2.add_patch(rectangle)
        # Arrow pointing to the left side of the rectangle
        ax2.arrow(
            0.5, 0.7, 0, 0.1, head_width=0.05, head_length=0.1, fc="black", ec="black"
        )
        # Arrow pointing out of the right side of the rectangle
        ax2.arrow(
            0.5, 0.1, 0.0, 0.1, head_width=0.05, head_length=0.1, fc="black", ec="black"
        )
        ax2.set_aspect("equal")
        ax2.set_xlim(0, 1)
        ax2.set_ylim(0, 1)
        if self.name is None:
            ax2.set_title("Component  ")
        else:
            ax2.set_title(self.name)

        # Add text over the arrows
        ax2.text(
            0.7,
            0.8,
            r"$T_o$=" + str(self.T_out) + " K",
            color="black",
            ha="center",
            va="center",
        )
        ax2.text(
            0.7,
            0.2,
            r"$T_i$=" + str(self.T_in) + " K",
            color="black",
            ha="center",
            va="center",
        )
        ax2.text(
            0.3,
            0.2,
            f"$c_i$={self.c_in:.4g} $mol/m^3$",
            color="black",
            ha="center",
            va="center",
        )
        ax2.text(
            0.5, 0.6, f"Q={self.Q/1E6:.3g} MW", color="black", ha="center", va="center"
        )
        ax2.text(
            0.5, 0.4, f"TBR={self.TBR:.3g}", color="black", ha="center", va="center"
        )

        ax2.text(
            0.3,
            0.8,
            rf"$c_o$={self.c_out:.4g}$mol/m^3$",
            color="black",
            ha="center",
            va="center",
        )
        ax2.axis("off")
        # Display the plot
        fig.tight_layout()
        return fig

    def get_flowrate(self):
        """
        Calculates the flow rate of the coolant in the breeding blanket component.
        """
        self.m_coolant = self.Q / ((self.T_out - self.T_in) * self.fluid.cp)
        return

    # def connect_to_component(self, component2: Union["Component", "BreedingBlanket"]):
    #     component2.update_attribute("c_in", self.c_out)
    def connect_to_component(self):
        return

    def get_cout(self, print_var: bool = False):
        """
        Calculates the outlet concentration of tritium in the breeding blanket component.

        Args:
            print_var (bool): If True, prints the intermediate variables.

        Returns:
            None
        """
        if self.m_coolant is None:
            self.get_flowrate()
        eV_to_J = physical_constants["electron volt-joule relationship"][0]
        reaction_energy = 17.6e6  # reaction energy in eV 17.6 MeV
        neutrons = self.Q / (reaction_energy * eV_to_J)

        tritium_gen = self.TBR * neutrons / N_A  ##moles
        if print_var:
            print("neu", neutrons)
            print("Trit", tritium_gen)
        if self.fluid.MS:
            self.c_out = (
                tritium_gen / 2 / (self.m_coolant / self.fluid.rho) + self.c_in
            )  ##concentration of Q2 in molten salts
        else:
            self.c_out = (
                tritium_gen / (self.m_coolant / self.fluid.rho) + self.c_in
            )  ## concentration of Q in liquid metals



import TRIOMA.tools.liquid_metals as lm


def W(k_r, D, thick, K_S, c0, K_S_L):
    """
    Calculates the value of W (surface/diffusion) using the given parameters.

    Parameters:
    k_r (float): The value of recombination coefficient.
    D (float): The value of Diffusivity in the metal.
    thick (float): The value of thickness of the metal.
    K_S (float): The value of the sievert's constant.
    c0 (float): The value of the initial concentration of hydrogen in the liquid metal.
    K_S_L (float): The value of the liquid sievert's constant.

    Returns:
    float: The calculated value of W.
    For W<<1, the regime is surface limited.
    flux is 0.5*k_d*P_H2^0.5
    """
    return k_r / D * K_S * thick * c0 / K_S_L


def partition_param(D, k_t, K_S_S, K_S_L, t):
    """
    Calculate the partition parameter Diffusion/mass transfer.

    Parameters:
    D (float): Diffusion coefficient.
    k_t (float): mass transfer coeff.
    K_S_S (float): Solid Sievert constant.
    K_S_L (float): Liquid Sievert constant.
    t (float): Thickness.

    Returns:
    float: The partition parameter.
    for partition<<1 membrane effects are slower than mass transfer
    when W<<1 surface limited
    # when W>>1 Diffusion limited limited
    if partition >>1 and W>1 mass transfer limited
    if partition >>1 and W<<1 surface limited and mass transfer limited
    """
    return D / k_t * K_S_S / (K_S_L * t)


def get_regime(D, k_t, K_S_S, K_S_L, k_r, thick, c0, print_var: bool = None):
    """
    Get the regime based on the value of H and W.
    D (float): Diffusion coefficient.
    k_t (float): mass transfer coeff.
    K_S_S (float): Solid Sievert constant.
    K_S_L (float): Liquid Sievert constant.
    k_r (float): Recombination coefficient.
    thick (float): Thickness.
    c0 (float): Initial concentration.

    Returns:
    str: The regime based on the value of H and W.
    """
    W = lm.W(k_r=k_r, D=D, thick=thick, K_S=K_S_S, c0=c0, K_S_L=K_S_L)
    partition_param = lm.partition_param(
        D=D, k_t=k_t, K_S_S=K_S_S, K_S_L=K_S_L, t=thick
    )
    result = None  ## initialize the result variable
    match (partition_param, W):
        case (partition_param, W) if partition_param > 10 and W > 10:
            result = "Mass transport limited"
        case (partition_param, W) if partition_param < 0.1 and W < 0.1:
            result = "Surface limited"
        case (partition_param, W) if W > 10 and partition_param < 0.1:
            result = "Diffusion Limited"
        case (partition_param, W) if partition_param > 10 and W < 0.1:
            result = "Transport and surface limited regime"
        case _:
            result = "Mixed regime"
    if print_var:
        print("H is equal to", partition_param * W)
        print("W is equal to", W)
        print("H/W is equal to", partition_param)
        print(result)
    return result
